from flask import Response, Blueprint
from platypush.plugins.camera import CameraPlugin

from platypush import Config
from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, send_request

camera = Blueprint('camera', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    camera,
]


def get_device_id(device_id=None):
    if device_id is None:
        device_id = int(send_request(action='camera.get_default_device_id').output)
    return device_id


def get_camera(device_id=None):
    device_id = get_device_id(device_id)
    camera_conf = Config.get('camera') or {}
    camera_conf['device_id'] = device_id
    return CameraPlugin(**camera_conf)


def get_frame(device_id=None):
    cam = get_camera(device_id)
    with cam:
        frame = None

        for _ in range(cam.warmup_frames):
            output = cam.get_stream()

            with output.ready:
                output.ready.wait()
                frame = output.frame

        return frame


def video_feed(device_id=None):
    cam = get_camera(device_id)

    with cam:
        while True:
            output = cam.get_stream()
            with output.ready:
                output.ready.wait()
                frame = output.frame

            if frame and len(frame):
                yield (b'--frame\r\n'
                       b'Content-Type: image/jpeg\r\n\r\n' + frame + b'\r\n')


@camera.route('/camera/<device_id>/frame', methods=['GET'])
@authenticate()
def get_camera_frame(device_id):
    frame = get_frame(device_id)
    return Response(frame, mimetype='image/jpeg')


@camera.route('/camera/frame', methods=['GET'])
@authenticate()
def get_default_camera_frame():
    frame = get_frame()
    return Response(frame, mimetype='image/jpeg')


@camera.route('/camera/<device_id>/stream', methods=['GET'])
@authenticate()
def get_stream_feed(device_id):
    return Response(video_feed(device_id),
                    mimetype='multipart/x-mixed-replace; boundary=frame')


@camera.route('/camera/stream', methods=['GET'])
@authenticate()
def get_default_stream_feed():
    return Response(video_feed(),
                    mimetype='multipart/x-mixed-replace; boundary=frame')


# vim:sw=4:ts=4:et:



import json
import requests

from platypush.config import Config
from platypush.message import Message
from platypush.plugins import Plugin, action


class AutoremotePlugin(Plugin):
    """
    This plugin allows you to send messages and notifications to an
    Android device that runs AutoRemote (https://joaoapps.com/autoremote/).
    You can also build custom actions to run on your Android device upon
    AutoRemote events using Tasker (https://tasker.joaoapps.com/).

    Requires:

        * **requests** (``pip install requests``)
    """

    _AUTOREMOTE_BASE_URL = 'https://autoremotejoaomgcd.appspot.com'
    _AUTOREMOTE_MESSAGE_URL = _AUTOREMOTE_BASE_URL + '/sendmessage'
    _AUTOREMOTE_NOTIFICATION_URL = _AUTOREMOTE_BASE_URL + '/sendnotification'

    def __init__(self, devices=None, key=None, password=None, *args, **kwargs):
        """
        :param devices: Set this attribute if you want to control multiple AutoRemote devices.
            This will be a map in the format::

                {
                    'device_name': {
                        'key': 'AUTOREMOTE_KEY',
                        'password': 'DEVICE_PASSWORD'
                    },
                    ...
                }

        :type devices: dict

        :param key: The key associated to your device. Open the link in your AutoRemote app
            and copy the key in the target URL. Set this value if you want to communicate
            with only one AutoRemote device.
        :type key: str

        :param password: AutoRemote password configured on the device (default: None).
            Set this value if you want to communicate with only one AutoRemote device.
        :type password: str
        """

        super().__init__(*args, **kwargs)

        if key:
            self.devices = { key: { 'key': key, 'password': password } }
        elif devices:
            self.devices = devices
        else:
            raise RuntimeError('Either <key,password> or devices attributes should be set')

    @action
    def send_message(self, msg, key=None, password=None, devices=None, target=None,
                     sender=None, ttl=None, group=None, *args, **kwargs):
        """
        Sends a message to AutoRemote.

        :param msg: Message to send

        :param key: Set it if you want to override the default devices
            (default: None, message sent to all the configured devices)
        :type key: str

        :param password: Set it if you want to override the default password (default: None)
        :type password: str

        :param devices: Set it if you want to send the message to a specific list of
            configured devices (default: None, message sent to all the configured devices)
        :type devices: list

        :param target: Message target (default: None)
        :type target: str

        :param sender: Message sender (default: None)
        :type target: str

        :param ttl: Message time-to-live in seconds (default: None)
        :type ttl: int

        :param group: Message group name (default: None)
        :type group: str
        """

        target_devices = []

        if devices:
            target_devices = [self.devices[name] for name in self.devices.keys()
                              if name in devices]
        elif key:
            device = { 'key': key }
            if password:
                device['password'] = password

            target_devices = [device]
        else:
            target_devices = self.devices.values()

        for device in target_devices:
            args = {
                'message': msg,
                'key': device['key'],
                'sender': sender or Config.get('device_id'),
            }

            if device.get('password'):
                args['password'] = device['password']

            if target: args['target'] = target
            if ttl: args['ttl'] = ttl
            if group: args['collapseKey'] = group

            response = requests.post(self._AUTOREMOTE_MESSAGE_URL, data=args)
            self.logger.info('Received response from AutoRemote: {}'.format(response.text))

    @action
    def send_notification(self, text=None, key=None, password=None, devices=None,
                          title=None, target=None, sender=None, ttl=None,
                          group=None, sound=None, vibration=None, url=None,
                          id=None, action=None, icon=None, led=None, ledon=None,
                          ledoff=None, picture=None, share=False, msg=None,
                          action1=None, action1_name=None, action1_icon=None,
                          action2=None, action2_name=None, action2_icon=None,
                          action3=None, action3_name=None, action3_icon=None,
                          persistent=False, statusbar_icon=None, ticker=None,
                          dismiss_on_touch=False, priority=0, number=None,
                          content_info=None, subtext=None, max_progress=None,
                          progress=None, indeterminate_progress=False,
                          action_on_dismiss=None, cancel=False, *args, **kwargs):
        """
        Sends a notification to AutoRemote. Click on your AutoRemote URL ->
        Send Notification for a detailed explanation of the attributes.
        """

        target_devices = []

        if devices:
            target_devices = [self.devices[name] for name in self.devices.keys()
                              if name in devices]
        elif key:
            device = { 'key': key }
            if password:
                device['password'] = password

            target_devices = [device]
        else:
            target_devices = self.devices.values()

        for device in target_devices:
            args = {
                'text': text,
                'key': device.get('key'),
                'sender': sender or Config.get('device_id'),
            }

            if device.get('password'):
                args['password'] = device['password']

            if target: args['target'] = target
            if ttl: args['ttl'] = ttl
            if group: args['collapseKey'] = group
            if title: args['title'] = title
            if text: args['text'] = text
            if sound: args['sound'] = sound
            if vibration: args['vibration'] = vibration
            if url: args['url'] = url
            if id: args['id'] = id
            if action: args['action'] = action
            if icon: args['icon'] = icon
            if led: args['led'] = led
            if ledon: args['ledon'] = ledon
            if ledoff: args['ledoff'] = ledoff
            if picture: args['picture'] = picture
            if msg: args['message'] = msg
            if share: args['share'] = share
            if action1: args['action1'] = action1
            if action1_name: args['action1name'] = action1_name
            if action1_icon: args['action1icon'] = action1_icon
            if action2: args['action2'] = action2
            if action2_name: args['action2name'] = action2_name
            if action2_icon: args['action2icon'] = action2_icon
            if action3: args['action3'] = action3
            if action3_name: args['action3name'] = action3_name
            if action3_icon: args['action3icon'] = action3_icon
            if persistent: args['persistent'] = persistent
            if statusbar_icon: args['statusbaricon'] = statusbar_icon
            if ticker: args['ticker'] = ticker
            if dismiss_on_touch: args['dismissontouch'] = dismiss_on_touch
            if priority: args['priority'] = priority
            if number: args['number'] = number
            if content_info: args['contentinfo'] = content_info
            if subtext: args['subtext'] = subtext
            if max_progress: args['maxprogress'] = max_progress
            if progress: args['progress'] = progress
            if indeterminate_progress: args['indeterminateprogress'] = indeterminate_progress
            if action_on_dismiss: args['actionondismiss'] = action_on_dismiss
            if cancel: args['cancel'] = cancel

            response = requests.post(self._AUTOREMOTE_NOTIFICATION_URL, data=args)
            self.logger.info('Received response from AutoRemote: {}'.format(response.text))


# vim:sw=4:ts=4:et:




import re
import subprocess

from platypush.context import get_plugin
from platypush.plugins.media import PlayerState

from platypush.plugins import Plugin, action

class MediaCtrlPlugin(Plugin):
    """
    Wrapper plugin to control audio and video media.
    Examples of supported URL types:

        - file:///media/movies/Movie.mp4 [requires media plugin enabled]
        - youtube:video:poAk9XgK7Cs [requires media plugin+youtube-dl]
        - magnet:?torrent_magnet [requires torrentcast]
        - spotify:track:track_id [leverages plugins.music.mpd]

    """

    _supported_plugins = {
        'music.mpd', 'media', 'video.torrentcast'
    }

    def __init__(self, torrentcast_port=9090, *args, **kwargs):
        super().__init__()
        self.torrentcast_port = torrentcast_port
        self.url = None
        self.plugin = None

    @classmethod
    def _get_type_and_resource_by_url(cls, url):
        # MPD/Mopidy media (TODO support more mopidy types)
        m = re.search('^https://open.spotify.com/([^?]+)', url)
        if m: url = 'spotify:{}'.format(m.group(1).replace('/', ':'))
        if url.startswith('spotify:') \
                or url.startswith('tunein:') \
                or url.startswith('soundcloud:'):
            return ('mpd', url)

        # YouTube video
        m = re.match('youtube:video:(.*)', url)
        if m: url = 'https://www.youtube.com/watch?v={}'.format(m.group(1))
        if url.startswith('https://www.youtube.com/watch?v='):
            return ('youtube:video', url)

        # Local media
        if url.startswith('file://'):
            m = re.match('^file://(.*)', url)
            return ('file', m.group(1))

        # URL to a .torrent media or Magnet link
        if url.startswith('magnet:') or url.endswith('.torrent'):
            return ('torrent', url)

        raise RuntimeError('Unknown URL type: {}'.format(url))


    def _get_playing_plugin(self):
        if self.plugin:
            status = self.plugin.status()
            if status['state'] == PlayerState.PLAY.value or state['state'] == PlayerState.PAUSE.value:
                return self.plugin

        for plugin in self._supported_plugins:
            try:
                player = get_plugin(plugin)
            except:
                try:
                    player = get_plugin(plugin, reload=True)
                except:
                    continue

            status = player.status().output
            if status['state'] == PlayerState.PLAY.value or status['state'] == PlayerState.PAUSE.value:
                return player

        return None


    @action
    def play(self, url):
        (type, resource) = self._get_type_and_resource_by_url(url)
        plugin_name = None

        if type == 'mpd':
            plugin_name = 'music.mpd'
        elif type == 'youtube:video' or type == 'file':
            plugin_name = 'media'
        elif type == 'torrent':
            plugin_name = 'video.torrentcast'

        if not plugin_name:
            raise RuntimeError("Unsupported type '{}'".format(type))

        try:
            self.plugin = get_plugin(plugin_name)
        except:
            self.plugin = get_plugin(plugin_name, reload=True)

        self.url = resource
        return self.plugin.play(resource)

    @action
    def pause(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.pause()


    @action
    def stop(self):
        plugin = self._get_playing_plugin()
        if plugin:
            ret = plugin.stop()
            self.plugin = None
            return ret


    @action
    def voldown(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.voldown()


    @action
    def volup(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.volup()


    @action
    def back(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.back()


    @action
    def forward(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.forward()


    @action
    def next(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.next()


    @action
    def previous(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.previous()


    @action
    def status(self):
        plugin = self._get_playing_plugin()
        if plugin: return plugin.status()


# vim:sw=4:ts=4:et:




import logging

from platypush.context import get_backend, get_plugin
from platypush.message.event import Event

logger = logging.getLogger(__name__)


class AssistantEvent(Event):
    """ Base class for assistant events """

    def __init__(self, assistant=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if assistant:
            self._assistant = assistant
        else:
            self._assistant = get_backend('assistant.google')

            if not self._assistant:
                self._assistant = get_plugin('assistant.google.pushtotalk')

            if not self._assistant:
                logger.warning('Assistant plugin/backend not configured/initialized')
                self._assistant = None


class ConversationStartEvent(AssistantEvent):
    """
    Event triggered when a new conversation starts
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ConversationEndEvent(AssistantEvent):
    """
    Event triggered when a conversation ends
    """

    def __init__(self, with_follow_on_turn=False, *args, **kwargs):
        """
        :param with_follow_on_turn: Set to true if the conversation expects a user follow-up, false otherwise
        :type with_follow_on_turn: str
        """

        super().__init__(*args, with_follow_on_turn=with_follow_on_turn, **kwargs)


class ConversationTimeoutEvent(ConversationEndEvent):
    """
    Event triggered when a conversation times out
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ResponseEvent(ConversationEndEvent):
    """
    Event triggered when a response is processed by the assistant
    """

    def __init__(self, response_text, *args, **kwargs):
        """
        :param response_text: Response text processed by the assistant
        :type response_text: str
        """

        super().__init__(*args, response_text=response_text, **kwargs)


class NoResponseEvent(ConversationEndEvent):
    """
    Event triggered when a conversation ends with no response
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SpeechRecognizedEvent(AssistantEvent):
    """
    Event triggered when a speech is recognized
    """

    def __init__(self, phrase, *args, **kwargs):
        """
        :param phrase: Recognized user phrase
        :type phrase: str
        """

        super().__init__(*args, phrase=phrase, **kwargs)
        self.recognized_phrase = phrase.strip().lower()

    def matches_condition(self, condition):
        """
        Overrides matches condition, and stops the conversation to prevent the
        default assistant response if the event matched some event hook condition
        """

        result = super().matches_condition(condition)
        if result.is_match and self._assistant and 'phrase' in condition.args:
            self._assistant.stop_conversation()

        return result


class HotwordDetectedEvent(AssistantEvent):
    """
    Event triggered when a custom hotword is detected
    """

    def __init__(self, hotword=None, *args, **kwargs):
        """
        :param hotword: The detected user hotword
        :type hotword: str
        """

        super().__init__(*args, hotword=hotword, **kwargs)


class VolumeChangedEvent(AssistantEvent):
    """
    Event triggered when the volume of the assistant changes
    """

    def __init__(self, volume, *args, **kwargs):
        super().__init__(*args, volume=volume, **kwargs)


class AlertStartedEvent(AssistantEvent):
    """
    Event triggered when an alert starts on the assistant
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AlertEndEvent(AssistantEvent):
    """
    Event triggered when an alert ends on the assistant
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AlarmStartedEvent(AlertStartedEvent):
    """
    Event triggered when an alarm starts on the assistant
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AlarmEndEvent(AlertEndEvent):
    """
    Event triggered when an alarm ends on the assistant
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TimerStartedEvent(AlertStartedEvent):
    """
    Event triggered when a timer starts on the assistant
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TimerEndEvent(AlertEndEvent):
    """
    Event triggered when a timer ends on the assistant
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MicMutedEvent(AssistantEvent):
    """
    Event triggered when the microphone is muted.
    """
    pass


class MicUnmutedEvent(AssistantEvent):
    """
    Event triggered when the microphone is muted.
    """
    pass


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message import Mapping


class EspWifiScanResult(Mapping):
    def __init__(self,
                 essid: str,
                 bssid: str,
                 channel: int,
                 rssi: int,
                 auth_mode: int,
                 hidden: bool,
                 *args,
                 **kwargs):
        self.essid = essid
        self.bssid = bssid
        self.channel = channel
        self.rssi = rssi
        self.auth_mode = auth_mode
        self.hidden = hidden
        super().__init__(*args, **dict(self), **kwargs)


class EspWifiConfigResult(Mapping):
    def __init__(self,
                 ip: str,
                 netmask: str,
                 gateway: str,
                 dns: str,
                 mac: str,
                 active: bool,
                 essid: Optional[str] = None,
                 channel: Optional[int] = None,
                 hidden: Optional[bool] = None,
                 *args,
                 **kwargs):
        self.ip = ip
        self.netmask = netmask
        self.gateway = gateway
        self.dns = dns
        self.mac = mac
        self.active = active
        self.essid = essid
        self.channel = channel
        self.hidden = hidden
        super().__init__(*args, **dict(self), **kwargs)


# vim:sw=4:ts=4:et:



import json

from typing import Optional, List

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.google.pubsub import GooglePubsubMessageEvent
from platypush.utils import set_thread_name


class GooglePubsubBackend(Backend):
    """
    Subscribe to a list of topics on a Google Pub/Sub instance. See
        :class:`platypush.plugins.google.pubsub.GooglePubsubPlugin` for a reference on how to generate your
        project and credentials file.

    Triggers:

        * :class:`platypush.message.event.google.pubsub.GooglePubsubMessageEvent` when a new message is received on
          a subscribed topic.

    Requires:

        * **google-cloud-pubsub** (``pip install google-cloud-pubsub``)

    """

    def __init__(self, topics: List[str], credentials_file: Optional[str] = None, *args, **kwargs):
        """
        :param topics: List of topics to subscribe. You can either specify the full topic name in the format
            ``projects/<project_id>/topics/<topic_name>``, where ``<project_id>`` must be the ID of your
            Google Pub/Sub project, or just ``<topic_name>``  - in such case it's implied that you refer to the
            ``topic_name`` under the ``project_id`` of your service credentials.
        :param credentials_file: Path to the Pub/Sub service credentials file (default: value configured on the
            ``google.pubsub`` plugin or ``~/.credentials/platypush/google/pubsub.json``).
        """

        super().__init__(*args, **kwargs)
        self.topics = topics

        if credentials_file:
            self.credentials_file = credentials_file
        else:
            plugin = self._get_plugin()
            self.credentials_file = plugin.credentials_file

    @staticmethod
    def _get_plugin():
        return get_plugin('google.pubsub')

    def _message_callback(self, topic):
        def callback(msg):
            data = msg.data.decode()
            # noinspection PyBroadException
            try:
                data = json.loads(data)
            except:
                pass

            msg.ack()
            self.bus.post(GooglePubsubMessageEvent(topic=topic, msg=data))

        return callback

    def run(self):
        # noinspection PyPackageRequirements
        from google.cloud import pubsub_v1
        # noinspection PyPackageRequirements
        from google.api_core.exceptions import AlreadyExists

        super().run()
        set_thread_name('GooglePubSub')
        plugin = self._get_plugin()
        project_id = plugin.get_project_id()
        credentials = plugin.get_credentials(plugin.subscriber_audience)
        subscriber = pubsub_v1.SubscriberClient(credentials=credentials)

        for topic in self.topics:
            if not topic.startswith('projects/{}/topics/'.format(project_id)):
                topic = 'projects/{}/topics/{}'.format(project_id, topic)
            subscription_name = '/'.join([*topic.split('/')[:2], 'subscriptions', topic.split('/')[-1]])

            try:
                subscriber.create_subscription(name=subscription_name, topic=topic)
            except AlreadyExists:
                pass

            subscriber.subscribe(subscription_name, self._message_callback(topic))

        self.wait_stop()


# vim:sw=4:ts=4:et:



import datetime

from typing import Optional, List, Union

from platypush.message import Mapping
from platypush.message.response import Response


class TrelloResponse(Response):
    pass


class TrelloList(Mapping):
    def __init__(self,
                 id: str,
                 name: str,
                 closed: bool,
                 subscribed: bool,
                 *args, **kwargs):
        super().__init__(id=id, name=name, closed=closed, subscribed=subscribed, *args, **kwargs)


class TrelloBoard(Mapping):
    def __init__(self,
                 id: str,
                 name: str,
                 url: str,
                 closed: bool,
                 lists: Optional[List[TrelloList]] = None,
                 description: Optional[str] = None,
                 date_last_activity: Optional[datetime.datetime] = None,
                 *args, **kwargs):
        super().__init__(id=id, name=name, url=url, closed=closed, description=description, lists=lists,
                         date_last_activity=date_last_activity, *args, **kwargs)


class TrelloBoardResponse(TrelloResponse):
    def __init__(self, board: TrelloBoard, **kwargs):
        super().__init__(output=board, **kwargs)
        self.board = board


class TrelloBoardsResponse(TrelloResponse):
    def __init__(self, boards: List[TrelloBoard], **kwargs):
        super().__init__(output=boards, **kwargs)


class TrelloListsResponse(TrelloResponse):
    def __init__(self, lists: List[TrelloList], **kwargs):
        super().__init__(output=lists, **kwargs)


class TrelloLabel(Mapping):
    def __init__(self,
                 id: str,
                 name: str,
                 color: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(id=id, name=name, color=color, *args, **kwargs)


class TrelloUser(Mapping):
    def __init__(self,
                 id: str,
                 username: str,
                 fullname: str,
                 initials: Optional[str] = None,
                 avatar_url: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(id=id, username=username, fullname=fullname, initials=initials,
                         avatar_url=avatar_url, *args, **kwargs)


class TrelloComment(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 id: str,
                 text: str,
                 type: str,
                 creator: TrelloUser,
                 date: Union[str, datetime.datetime],
                 *args, **kwargs):
        super().__init__(id=id, text=text, type=type, creator=creator, date=date, *args, **kwargs)


class TrelloCard(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 id: str,
                 name: str,
                 url: str,
                 closed: bool,
                 board: TrelloBoard,
                 is_due_complete: bool,
                 list: Optional[TrelloList] = None,
                 comments: Optional[List[TrelloComment]] = None,
                 labels: Optional[List[TrelloLabel]] = None,
                 description: Optional[str] = None,
                 due_date: Optional[Union[datetime.datetime, str]] = None,
                 latest_card_move_date: Optional[Union[datetime.datetime, str]] = None,
                 date_last_activity: Optional[Union[datetime.datetime, str]] = None,
                 *args, **kwargs):
        super().__init__(id=id, name=name, url=url, closed=closed, board=board, is_due_complete=is_due_complete,
                         description=description, date_last_activity=date_last_activity, due_date=due_date, list=list,
                         comments=comments, labels=labels, latest_card_move_date=latest_card_move_date, *args, **kwargs)


class TrelloCardResponse(TrelloResponse):
    def __init__(self, card: TrelloCard, **kwargs):
        super().__init__(output=card, **kwargs)


class TrelloCardsResponse(TrelloResponse):
    def __init__(self, cards: List[TrelloCard], **kwargs):
        super().__init__(output=cards, **kwargs)


class TrelloPreview(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 id: str,
                 scaled: bool,
                 url: str,
                 bytes: int,
                 height: int,
                 width: int,
                 *args, **kwargs):
        super().__init__(id=id, scaled=scaled, url=url, bytes=bytes, height=height, width=width, *args, **kwargs)


class TrelloAttachment(Mapping):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 id: str,
                 bytes: int,
                 date: str,
                 edge_color: str,
                 id_member: str,
                 is_upload: bool,
                 name: str,
                 previews: List[TrelloPreview],
                 url: str,
                 mime_type: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(id=id, bytes=bytes, date=date, edge_color=edge_color, id_member=id_member, is_upload=is_upload,
                         name=name, previews=previews, url=url, mime_type=mime_type, *args, **kwargs)


class TrelloChecklistItem(Mapping):
    def __init__(self,
                 id: str,
                 name: str,
                 checked: bool,
                 *args, **kwargs):
        super().__init__(id=id, name=name, checked=checked, *args, **kwargs)


class TrelloChecklist(Mapping):
    def __init__(self,
                 id: str,
                 name: str,
                 checklist_items: List[TrelloChecklistItem],
                 *args, **kwargs):
        super().__init__(id=id, name=name, checklist_items=checklist_items, *args, **kwargs)


class TrelloMember(Mapping):
    def __init__(self,
                 id: str,
                 full_name: str,
                 bio: Optional[str],
                 url: Optional[str],
                 username: Optional[str],
                 initials: Optional[str],
                 member_type: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(id=id, full_name=full_name, bio=bio, url=url, username=username, initials=initials,
                         member_type=member_type, *args, **kwargs)


class TrelloMembersResponse(Mapping):
    def __init__(self, members: List[TrelloMember], **kwargs):
        super().__init__(output=members, **kwargs)


# vim:sw=4:ts=4:et:



import time

from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorDistanceVl53L1XPlugin(GpioSensorPlugin):
    """
    Plugin to interact with an `VL53L1x <https://www.st.com/en/imaging-and-photonics-solutions/vl53l1x.html>`_
    laser ranger/distance sensor

    Requires:

        * ``smbus2`` (``pip install smbus2``)
        * ``vl53l1x`` (``pip install vl53l1x``)
    """

    def __init__(self, i2c_bus=1, i2c_address=0x29, **kwargs):
        """
        :param i2c_bus: I2C bus number (default: 1)
        :param i2c_address: I2C address (default: 0x29)
        """

        super().__init__(**kwargs)

        self.i2c_bus = i2c_bus
        self.i2c_address = i2c_address
        self._device = None

    # noinspection PyUnresolvedReferences
    def _get_device(self, ranging=1):
        if self._device:
            return self._device

        from VL53L1X import VL53L1X
        self._device = VL53L1X(i2c_bus=self.i2c_bus, i2c_address=self.i2c_address)
        self._device.open()
        self._device.start_ranging(ranging)
        return self._device

    @action
    def get_measurement(self, short=True, medium=False, long=False):
        """
        :param short: Enable short range measurement (default: True)
        :param medium: Enable medium range measurement (default: False)
        :param long: Enable long range measurement (default: False)

        :returns: dict. Example:

        .. code-block:: python

            output = {
                "short": 83,     # Short range measurement in mm
                "medium": 103,   # Medium range measurement
                "long": 43,      # Long range measurement
            }

        """

        range_idx = 0
        range_name = None

        for i, r in enumerate(['short', 'medium', 'long']):
            if eval(r):
                range_idx = i+1
                range_name = r
                break

        assert range_name is not None
        device = self._get_device(ranging=range_idx)
        ret = {}

        try:
            ret[range_name] = device.get_distance()
        except Exception as e:
            self.logger.exception(e)

            try:
                self._device.stop_ranging()
                self._device.close()
            except:
                pass

            self._device = None
            time.sleep(1)

        return ret


# vim:sw=4:ts=4:et:



import hashlib
import json
import threading

from flask import Response, render_template

from platypush.backend.http.app.utils import get_remote_base_url, logger, \
    send_message

from platypush.backend.http.media.handlers import MediaHandler

media_map = {}
media_map_lock = threading.RLock()

# Size for the bytes chunk sent over the media streaming infra
STREAMING_CHUNK_SIZE = 4096

# Maximum range size to be sent through the media streamer if Range header
# is not set
STREAMING_BLOCK_SIZE = 3145728


def get_media_url(media_id):
    return '{url}/media/{media_id}'.format(
        url=get_remote_base_url(), media_id=media_id)

def get_media_id(source):
    return hashlib.sha1(source.encode()).hexdigest()

def register_media(source, subtitles=None):
    global media_map, media_map_lock

    media_id = get_media_id(source)
    media_url = get_media_url(media_id)

    with media_map_lock:
        if media_id in media_map:
            return media_map[media_id]

    subfile = None
    if subtitles:
        req = {
            'type': 'request',
            'action': 'media.subtitles.download',
            'args': {
                'link': subtitles,
                'convert_to_vtt': True,
            }
        }

        try:
            subfile = (send_message(req).output or {}).get('filename')
        except Exception as e:
            logger().warning('Unable to load subtitle {}: {}'
                             .format(subtitles, str(e)))

    with media_map_lock:
        media_hndl = MediaHandler.build(source, url=media_url, subtitles=subfile)
        media_map[media_id] = media_hndl
        media_hndl.media_id = media_id

    logger().info('Streaming "{}" on {}'.format(source, media_url))
    return media_hndl


def unregister_media(source):
    global media_map, media_map_lock

    if source is None:
        raise KeyError('No media_id specified')

    media_id = get_media_id(source)
    media_info = {}

    with media_map_lock:
        if media_id not in media_map:
            raise FileNotFoundError('{} is not a registered media_id'.
                                    format(source))
        media_info = media_map.pop(media_id)

    logger().info('Unregistered {} from {}'.format(source, media_info.get('url')))
    return media_info


def stream_media(media_id, req):
    global STREAMING_BLOCK_SIZE, STREAMING_CHUNK_SIZE

    media_hndl = media_map.get(media_id)
    if not media_hndl:
        raise FileNotFoundError('{} is not a registered media_id'.format(media_id))

    from_bytes = None
    to_bytes = None
    range_hdr = req.headers.get('range')
    content_length = media_hndl.content_length
    status_code = 200

    headers = {
        'Accept-Ranges': 'bytes',
        'Content-Type': media_hndl.mime_type,
    }

    if 'download' in req.args:
        headers['Content-Disposition'] = 'attachment' + \
            ('; filename="{}"'.format(media_hndl.filename) if
             media_hndl.filename else '')

    if range_hdr:
        headers['Accept-Ranges'] = 'bytes'
        from_bytes, to_bytes = range_hdr.replace('bytes=', '').split('-')
        from_bytes = int(from_bytes)

        if not to_bytes:
            to_bytes = content_length-1
            content_length -= from_bytes
        else:
            to_bytes = int(to_bytes)
            content_length = to_bytes - from_bytes

        status_code = 206
        headers['Content-Range'] = 'bytes {start}-{end}/{size}'.format(
            start=from_bytes, end=to_bytes,
            size=media_hndl.content_length)
    else:
        from_bytes = 0
        to_bytes = STREAMING_BLOCK_SIZE

    headers['Content-Length'] = content_length

    if 'webplayer' in req.args:
        return render_template('webplayer.html',
                                media_url=media_hndl.url.replace(
                                    get_remote_base_url(), ''),
                                media_type=media_hndl.mime_type,
                                subtitles_url='/media/subtitles/{}.vtt'.
                                format(media_id) if media_hndl.subtitles
                                else None)
    else:
        return Response(media_hndl.get_data(
            from_bytes=from_bytes, to_bytes=to_bytes,
            chunk_size=STREAMING_CHUNK_SIZE),
            status_code, headers=headers, mimetype=headers['Content-Type'],
            direct_passthrough=True)


def add_subtitles(media_id, req):
    """
    This route can be used to download and/or expose subtitles files
    associated to a media file
    """

    media_hndl = media_map.get(media_id)
    if not media_hndl:
        raise FileNotFoundError('{} is not a registered media_id'.format(media_id))

    subfile = None
    if req.data:
        subfile = json.loads(req.data.decode('utf-8')).get('filename')
        if not subfile:
            raise AttributeError('No filename specified in the request')

    if not subfile:
        if not media_hndl.path:
            raise NotImplementedError(
                'Subtitles are currently only supported for local media files')

        req = {
            'type': 'request',
            'action': 'media.subtitles.get_subtitles',
            'args': {
                'resource': media_hndl.path,
            }
        }

        try:
            subtitles = send_message(req).output or []
        except Exception as e:
            raise RuntimeError('Could not get subtitles: {}'.format(str(e)))

        if not subtitles:
            raise FileNotFoundError('No subtitles found for resource {}'.
                                    format(media_hndl.path))

        req = {
            'type': 'request',
            'action': 'media.subtitles.download',
            'args': {
                'link': subtitles[0].get('SubDownloadLink'),
                'media_resource': media_hndl.path,
                'convert_to_vtt': True,
            }
        }

        subfile = (send_message(req).output or {}).get('filename')

    media_hndl.set_subtitles(subfile)
    return {
        'filename': subfile,
        'url': get_remote_base_url() + '/media/subtitles/' + media_id + '.vtt',
    }

def remove_subtitles(media_id):
    media_hndl = media_map.get(media_id)
    if not media_hndl:
        raise FileNotFoundError('{} is not a registered media_id'.
                                format(media_id))

    if not media_hndl.subtitles:
        raise FileNotFoundError('{} has no subtitles attached'.
                                format(media_id))

    media_hndl.remove_subtitles()
    return {}



# vim:sw=4:ts=4:et:



import os
import tempfile

from flask import Response, Blueprint, request

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate, send_request

sound = Blueprint('sound', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    sound,
]


# Generates the .wav file header for a given set of samples and specs
# noinspection PyRedundantParentheses
def gen_header(sample_rate, sample_width, channels):
    datasize = int(2000 * 1e6)                                               # Arbitrary data size for streaming
    o = bytes("RIFF", ' ascii')                                               # (4byte) Marks file as RIFF
    o += (datasize + 36).to_bytes(4, 'little')                               # (4byte) File size in bytes
    o += bytes("WAVE", 'ascii')                                              # (4byte) File type
    o += bytes("fmt ", 'ascii')                                              # (4byte) Format Chunk Marker
    o += (16).to_bytes(4, 'little')                                          # (4byte) Length of above format data
    o += (1).to_bytes(2, 'little')                                           # (2byte) Format type (1 - PCM)
    o += channels.to_bytes(2, 'little')                                      # (2byte)
    o += sample_rate.to_bytes(4,  'little')                                  # (4byte)
    o += (sample_rate * channels * sample_width // 8).to_bytes(4, 'little')  # (4byte)
    o += (channels * sample_width // 8).to_bytes(2, 'little')                # (2byte)
    o += sample_width.to_bytes(2, 'little')                                  # (2byte)
    o += bytes("data", 'ascii')                                              # (4byte) Data Chunk Marker
    o += datasize.to_bytes(4, 'little')                                      # (4byte) Data size in bytes
    return o


def audio_feed(device, fifo, sample_rate, blocksize, latency, channels):
    send_request(action='sound.stream_recording', device=device, sample_rate=sample_rate,
                 dtype='int16', fifo=fifo, blocksize=blocksize, latency=latency,
                 channels=channels)

    try:
        with open(fifo, 'rb') as f:
            send_header = True

            while True:
                audio = f.read(blocksize)

                if audio:
                    if send_header:
                        audio = gen_header(sample_rate=sample_rate, sample_width=16, channels=channels) + audio
                        send_header = False

                    yield audio
    finally:
        send_request(action='sound.stop_recording')


@sound.route('/sound/stream', methods=['GET'])
@authenticate()
def get_sound_feed():
    device = request.args.get('device')
    sample_rate = request.args.get('sample_rate', 44100)
    blocksize = request.args.get('blocksize', 512)
    latency = request.args.get('latency', 0)
    channels = request.args.get('channels', 1)
    fifo = request.args.get('fifo', os.path.join(tempfile.gettempdir(), 'inputstream'))

    return Response(audio_feed(device=device, fifo=fifo, sample_rate=sample_rate,
                               blocksize=blocksize, latency=latency, channels=channels),
                    mimetype='audio/x-wav;codec=pcm')


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import datetime
import dateutil.parser
import requests

from platypush.plugins import Plugin, action
from platypush.plugins.calendar import CalendarInterface


class CalendarIcalPlugin(Plugin, CalendarInterface):
    """
    iCal calendars plugin. Interact with remote calendars in iCal format.

    Requires:

        * **icalendar** (``pip install icalendar``)
        * **python-dateutil** (``pip install python-dateutil``)
    """

    def __init__(self, url, *args, **kwargs):
        """
        :param url: iCal URL to parse
        :type url: str
        """
        super().__init__(*args, **kwargs)
        self.url = url


    def _translate_event(self, event):
        return {
            'id': str(event.get('uid')) if event.get('uid') else None,
            'kind': 'calendar#event',
            'summary': str(event.get('summary')) if event.get('summary') else None,
            'description': str(event.get('description')) if event.get('description') else None,
            'status': str(event.get('status')).lower() if event.get('status') else None,
            'responseStatus': str(event.get('partstat')).lower() if event.get('partstat') else None,
            'location': str(event.get('location')) if event.get('location') else None,
            'htmlLink': str(event.get('url')) if event.get('url') else None,
            'organizer': {
                'email': str(event.get('organizer')).replace('MAILTO:', ''),
                'displayName': event.get('organizer').params['cn']
            } if event.get('organizer') else None,

            'created': event.get('created').dt.isoformat() if event.get('created') else None,
            'updated': event.get('last-modified').dt.isoformat() if event.get('last-modified') else None,
            'start': {
                'dateTime': event.get('dtstart').dt.isoformat() if event.get('dtstart') else None,
                'timeZone': 'UTC',  # TODO Support multiple timezone IDs
            },

            'end': {
                'dateTime': event.get('dtend').dt.isoformat() if event.get('dtend') else None,
                'timeZone': 'UTC',
            },
        }


    @action
    def get_upcoming_events(self, max_results=10, only_participating=True):
        """
        Get the upcoming events. See
        :func:`~platypush.plugins.calendar.CalendarPlugin.get_upcoming_events`.
        """

        import pytz
        from icalendar import Calendar

        events = []
        try:
            response = requests.get(self.url)
        except Exception as e:
            self.logger.exception(e)
            return events

        if response.ok:
            calendar = None

            try:
                calendar = Calendar.from_ical(response.text)
            except Exception as e:
                self.logger.exception(e)
                return events

            for event in calendar.walk():
                if event.name != 'VEVENT':
                    continue  # Not an event

                event = self._translate_event(event)

                if event['status'] and event['responseStatus'] \
                        and dateutil.parser.parse(event['end']['dateTime']) >= \
                            datetime.datetime.now(pytz.timezone('UTC')) \
                        and (
                            (only_participating
                             and event['status'] == 'confirmed'
                             and event['responseStatus'] in ['accepted', 'tentative'])
                            or not only_participating):
                    events.append(event)
        else:
            self.logger.error("HTTP error while getting {}: {}".format(self.url, response))

        return events


# vim:sw=4:ts=4:et:




import datetime
import json
import os
import subprocess

from platypush.plugins import action
from platypush.plugins.http.request import Plugin


class HttpWebpagePlugin(Plugin):
    """
    Plugin to handle and parse/simplify web pages.
    It used to use the Mercury Reader web API, but now that the API is discontinued this plugin is basically a
    wrapper around the `mercury-parser <https://github.com/postlight/mercury-parser>`_ JavaScript library.

    Requires:

        * **requests** (``pip install requests``)
        * **weasyprint** (``pip install weasyprint``), optional, for HTML->PDF conversion
        * **node** and **npm** installed on your system (to use the mercury-parser interface)
        * The mercury-parser library installed (``npm install @postlight/mercury-parser``)
    """

    _mercury_script = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mercury-parser.js')

    @action
    def simplify(self, url, outfile=None):
        """
        Parse the content of a web page removing any extra elements using Mercury

        :param url: URL to parse
        :param outfile: If set then the output will be written to the specified file
            (supported formats: pdf, html, plain (default)). The plugin will guess
            the format from the extension
        :return: dict

        Example if outfile is not specified::

            {
                "url": <url>,
                "title": <page title>,
                "content": <page parsed content>

            }

        Example if outfile is specified::

            {
                "url": <url>,
                "title": <page title>,
                "outfile": <output file absolute path>

            }

        """

        self.logger.info('Parsing URL {}'.format(url))
        parser = subprocess.Popen(['node', self._mercury_script, url], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        response = parser.stdout.read().decode()

        try:
            response = json.loads(response)
        except Exception as e:
            raise RuntimeError('Could not parse JSON: {}. Response: {}'.format(str(e), response))

        self.logger.info('Got response from Mercury API: {}'.format(response))
        title = response.get('title', '{} on {}'.format(
            'Published' if response.get('date_published') else 'Generated',
            response.get('date_published', datetime.datetime.now().isoformat())))

        content = response.get('content', '')

        if not outfile:
            return {
                'url': url,
                'title': title,
                'content': content,
            }

        outfile = os.path.abspath(os.path.expanduser(outfile))
        content = '''<h1>{title}</h1><div class="_parsed-content">{content}</div>'''.\
            format(title=title, content=content)

        style = '''
            body {
                font-size: 22px;
                font-family: 'Merriweather', Georgia, 'Times New Roman', Times, serif;
            }
            '''

        if outfile.lower().endswith('.pdf'):
            import weasyprint
            from weasyprint.fonts import FontConfiguration

            font_config = FontConfiguration()
            css = [weasyprint.CSS('https://fonts.googleapis.com/css?family=Merriweather'),
                   weasyprint.CSS(string=style, font_config=font_config)]

            weasyprint.HTML(string=content).write_pdf(outfile, stylesheets=css)
        else:
            content = '''
                <html>
                    <head>
                        <title>{title}</title>
                        <style>{style}</style>
                    </head>
                    <body>{{content}}</body>
                </html>
            '''.format(title=title, style=style, content=content)

            with open(outfile, 'w', encoding='utf-8') as f:
                f.write(content)

        return {
            'url': url,
            'title': title,
            'outfile': outfile,
        }


# vim:sw=4:ts=4:et:



import requests
from typing import Callable, Dict, Any, List, Optional

from platypush.plugins import Plugin, action


class TravisciPlugin(Plugin):
    """
    `Travis-Ci <https://travis-ci.org/>`_ continuous integration plugin.

    Setup:

        - Get your API token from your `Travis-Ci account settings page <https://travis-ci.org/account/preferences>`_.

    """

    api_base_url = 'https://api.travis-ci.org/'

    def __init__(self, token: str, **kwargs):
        super().__init__(**kwargs)
        self.headers = {
            'Travis-API-Version': '3',
            'Authorization':  'token ' + token,
        }

    def _make_request(self, method: Callable, endpoint: str, **kwargs):
        url = self.api_base_url + endpoint
        response = method(url, headers=self.headers, **kwargs).json()
        if response.get('@type') == 'error':
            raise AssertionError('{type}: {message}'.
                                 format(type=response.get('error_type'), message=response.get('error_message')))

        return response

    @action
    def repos(self) -> Dict[str, Dict[str, Any]]:
        """
        Get the repos owned by current user.
        :return: Repo name -> Repo attributes mapping.
        """
        return {
            repo['name']: repo
            for repo in self._make_request(requests.get, endpoint='repos').get('repositories', [])
        }

    @action
    def builds(self, limit: int = 100) -> Dict[str, List[Dict[str, Any]]]:
        """
        Get the list of builds triggered on the owned repositories

        :param limit: Maximum number of builds to be retrieved (default: 100).
        :return: Repo name -> List of builds
        """
        return self._make_request(requests.get, endpoint='builds').get('builds', [])[:limit]


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class MidiMessageEvent(Event):
    """
    Event triggered upon received MIDI message
    """

    def __init__(self, message, delay=None, *args, **kwargs):
        """
        :param message: Received MIDI message
        :type message: tuple[int]

        :param delay: Time in seconds since the previous MIDI event (default: None)
        :type delay: float
        """

        super().__init__(*args, message=message, delay=delay, **kwargs)


# vim:sw=4:ts=4:et:




"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import enum
import json
import math


class WaveShape(enum.Enum):
    SIN='sin'
    SQUARE='square'
    SAWTOOTH='sawtooth'
    TRIANG='triang'


class Sound(object):
    """
    Models a basic synthetic sound that can be played through an audio device
    """

    STANDARD_A_FREQUENCY = 440.0
    STANDARD_A_MIDI_NOTE = 69
    _DEFAULT_BLOCKSIZE = 1024
    _DEFAULT_SYNTH_BUFSIZE = 2
    _DEFAULT_FILE_BUFSIZE = 20
    _DEFAULT_SAMPLERATE = 44100

    midi_note = None
    frequency = None
    phase = 0.0
    gain = 1.0
    duration = None
    shape = None

    def __init__(self, midi_note=midi_note, frequency=None, phase=phase,
                 gain=gain, duration=duration, shape=WaveShape.SIN,
                 A_frequency=STANDARD_A_FREQUENCY):
        """
        You can construct a sound either from a MIDI note or a base frequency

        :param midi_note: MIDI note code, see
            https://newt.phys.unsw.edu.au/jw/graphics/notes.GIF
        :type midi_note: int

        :param frequency: Sound base frequency in Hz
        :type frequency: float

        :param phase: Wave phase shift as a multiple of pi (default: 0.0)
        :type phase: float

        :param gain: Note gain/volume between 0.0 and 1.0 (default: 1.0)
        :type gain: float

        :param duration: Note duration in seconds. Default: keep until
            release/pause/stop
        :type duration: float

        :param shape: Wave shape. Possible values: "``sin``", "``square``",
            "``sawtooth``" or "``triang``" (see :class:`WaveSound`).
            Default: "``sin``"
        :type shape: str

        :param A_frequency: Reference A4 frequency (default: 440 Hz)
        :type A_frequency: float
        """

        if midi_note and frequency:
            raise RuntimeError('Please specify either a MIDI note or a base ' +
                               'frequency')

        if midi_note:
            self.midi_note = midi_note
            self.frequency = self.note_to_freq(midi_note=midi_note,
                                               A_frequency=A_frequency)
        elif frequency:
            self.frequency = frequency
            self.midi_note = self.freq_to_note(frequency=frequency,
                                               A_frequency=A_frequency)
        else:
            raise RuntimeError('Please specify either a MIDI note or a base ' +
                               'frequency')

        self.phase = phase
        self.gain = gain
        self.duration = duration
        self.shape = WaveShape(shape)

    @classmethod
    def note_to_freq(cls, midi_note, A_frequency=STANDARD_A_FREQUENCY):
        """
        Converts a MIDI note to its frequency in Hz

        :param midi_note: MIDI note to convert
        :type midi_note: int

        :param A_frequency: Reference A4 frequency (default: 440 Hz)
        :type A_frequency: float
        """

        return (2.0 ** ((midi_note - cls.STANDARD_A_MIDI_NOTE) / 12.0)) \
            * A_frequency

    @classmethod
    def freq_to_note(cls, frequency, A_frequency=STANDARD_A_FREQUENCY):
        """
        Converts a frequency in Hz to its closest MIDI note

        :param frequency: Frequency in Hz
        :type midi_note: float

        :param A_frequency: Reference A4 frequency (default: 440 Hz)
        :type A_frequency: float
        """

        # TODO return also the offset in % between the provided frequency
        # and the standard MIDI note frequency
        return int(12.0 * math.log(frequency/A_frequency, 2)
                   + cls.STANDARD_A_MIDI_NOTE)

    def get_wave(self, t_start=0., t_end=0., samplerate=_DEFAULT_SAMPLERATE):
        """
        Get the wave binary data associated to this sound

        :param t_start: Start offset for the wave in seconds. Default: 0
        :type t_start: float

        :param t_end: End offset for the wave in seconds. Default: 0
        :type t_end: float

        :param samplerate: Audio sample rate. Default: 44100 Hz
        :type samplerate: int

        :returns: A ``numpy.ndarray[(t_end-t_start)*samplerate, 1]``
            with the raw float values
        """

        import numpy as np
        x = np.linspace(t_start, t_end, int((t_end-t_start)*samplerate))

        x = x.reshape(len(x), 1)

        if self.shape == WaveShape.SIN or self.shape == WaveShape.SQUARE:
            wave = np.sin((2*np.pi*self.frequency*x) + np.pi*self.phase)

            if self.shape == WaveShape.SQUARE:
                wave[wave < 0] = -1
                wave[wave >= 0] = 1
        elif self.shape == WaveShape.SAWTOOTH or self.shape == WaveShape.TRIANG:
            wave = 2 * (self.frequency*x -
                        np.floor(0.5 + self.frequency*x))
            if self.shape == WaveShape.TRIANG:
                wave = 2 * np.abs(wave) - 1
        else:
            raise RuntimeError('Unsupported wave shape: {}'.format(self.shape))

        return self.gain * wave


    def fft(self, t_start=0., t_end=0., samplerate=_DEFAULT_SAMPLERATE,
            freq_range=None, freq_buckets=None):
        """
        Get the real part of the Fourier transform associated to a time-bounded
            sample of this sound

        :param t_start: Start offset for the wave in seconds. Default: 0
        :type t_start: float

        :param t_end: End offset for the wave in seconds. Default: 0
        :type t_end: float

        :param samplerate: Audio sample rate. Default: 44100 Hz
        :type samplerate: int

        :param freq_range: FFT frequency range. Default: ``(0, samplerate/2)``
            (see `Nyquist-Shannon sampling theorem <https://en.wikipedia.org/wiki/Nyquist%E2%80%93Shannon_sampling_theorem>`_)
        :type freq_range: list or tuple with 2 int elements (range)

        :param freq_buckets: Number of buckets to subdivide the frequency range.
            Default: None
        :type freq_buckets: int

        :returns: A numpy.ndarray[freq_range,1] with the raw float values
        """

        import numpy as np

        if not freq_range:
            freq_range = (0, int(samplerate/2))

        wave = self.get_wave(t_start=t_start, t_end=t_end, samplerate=samplerate)
        fft = np.fft.fft(wave.reshape(len(wave)))
        fft = fft.real[freq_range[0]:freq_range[1]]

        if freq_buckets is not None:
            fft = np.histogram(fft, bins=freq_buckets)

        return fft

    def __iter__(self):
        for attr in ['midi_note', 'frequency', 'gain', 'duration']:
            yield (attr, getattr(self, attr))


    def __str__(self):
        return json.dumps(dict(self))


    @classmethod
    def build(cls, *args, **kwargs):
        """
        Construct a sound object either from a JSON representation or a
        key-value representation
        """

        if args:
            if isinstance(args[0], cls):
                return args[0]
            if isinstance(args[0], str):
                kwargs = json.loads(args[0])
            elif isinstance(args[0], dict):
                kwargs = args[0]
        if kwargs:
            return Sound(**kwargs)

        raise RuntimeError('Usage: {}'.format(__doc__))


class Mix(object):
    """
    This class models a set of mixed :class:`Sound` instances that can be played
    through an audio stream to an audio device
    """

    _sounds = None

    def __init__(self, *sounds):
        self._sounds = []

        for sound in sounds:
            self.add(sound)


    def __iter__(self):
        for sound in self._sounds:
            yield dict(sound)


    def __str__(self):
        return json.dumps(list(self))


    def add(self, sound):
        self._sounds.append(Sound.build(sound))


    def remove(self, sound_index):
        if sound_index >= len(self._sounds):
            self.logger.error('No such sound index: {} in mix {}'.format(
                sound_index, list(self)))
            return

        self._sounds.pop(sound_index)


    def get_wave(self, t_start=0., t_end=0., normalize_range=(-1.0, 1.0),
                 on_clip='scale', samplerate=Sound._DEFAULT_SAMPLERATE):
        """
        Get the wave binary data associated to this mix

        :param t_start: Start offset for the wave in seconds. Default: 0
        :type t_start: float

        :param t_end: End offset for the wave in seconds. Default: 0
        :type t_end: float

        :param normalize_range: Normalization range. If set the gain values of the
            wave will be normalized to fit into the specified range if it
            "clips" above or below.  Default: ``(-1.0, 1.0)``
        :type normalize_range: list[float]

        :param on_clip: Action to take on wave clipping if ``normalize_range``
            is set. Possible values: "``scale``" (scale down the frame to remove
            the clipping) or "``clip``" (saturate the values above/below range).
            Default: "``scale``".
        :type on_clip: str

        :param samplerate: Audio sample rate. Default: 44100 Hz
        :type samplerate: int

        :returns: A numpy.ndarray[n,1] with the raw float values
        """

        wave = None

        for sound in self._sounds:
            sound_wave = sound.get_wave(t_start=t_start, t_end=t_end,
                                        samplerate=samplerate)

            if wave is None:
                wave = sound_wave
            else:
                wave += sound_wave

        if normalize_range and len(wave):
            scale_factor = (normalize_range[1]-normalize_range[0]) / \
                (wave.max()-wave.min())

            if scale_factor < 1.0:  # Wave clipping
                if on_clip == 'scale':
                    wave = scale_factor * wave
                elif on_clip == 'clip':
                    wave[wave < normalize_range[0]] = normalize_range[0]
                    wave[wave > normalize_range[1]] = normalize_range[1]
                else:
                    raise RuntimeError('Supported values for "on_clip": ' +
                                       '"scale" or "clip"')

        return wave


    def fft(self, t_start=0., t_end=0., samplerate=Sound._DEFAULT_SAMPLERATE,
            freq_range=None, freq_buckets=None):
        """
        Get the real part of the Fourier transform associated to a time-bounded
            sample of this mix

        :param t_start: Start offset for the wave in seconds. Default: 0
        :type t_start: float

        :param t_end: End offset for the wave in seconds. Default: 0
        :type t_end: float

        :param samplerate: Audio sample rate. Default: 44100 Hz
        :type samplerate: int

        :param freq_range: FFT frequency range. Default: ``(0, samplerate/2)``
            (see `Nyquist-Shannon sampling theorem <https://en.wikipedia.org/wiki/Nyquist%E2%80%93Shannon_sampling_theorem>`_)
        :type freq_range: list or tuple with 2 int elements (range)

        :param freq_buckets: Number of buckets to subdivide the frequency range.
            Default: None
        :type freq_buckets: int

        :returns: A numpy.ndarray[freq_range,1] with the raw float values
        """

        import numpy as np

        if not freq_range:
            freq_range = (0, int(samplerate/2))

        wave = self.get_wave(t_start=t_start, t_end=t_end, samplerate=samplerate)
        fft = np.fft.fft(wave.reshape(len(wave)))
        fft = fft.real[freq_range[0]:freq_range[1]]

        if freq_buckets is not None:
            fft = np.histogram(fft, bins=freq_buckets)

        return fft


    def duration(self):
        """
        :returns: The duration of the mix in seconds as duration of its longest
            sample, or None if the mixed sample have no duration set
        """

        duration = 0

        for sound in self._sounds:
            if sound.duration is None:
                return None

            duration = max(duration, sound.duration)

        return duration


# vim:sw=4:ts=4:et:



import socket

from datetime import datetime
from typing import Union, List, Optional, Dict

from platypush.message.response.system import CpuInfoResponse, CpuTimesResponse, CpuResponseList, CpuStatsResponse, \
    CpuFrequencyResponse, VirtualMemoryUsageResponse, SwapMemoryUsageResponse, DiskResponseList, \
    DiskPartitionResponse, DiskUsageResponse, DiskIoCountersResponse, NetworkIoCountersResponse, NetworkResponseList, \
    NetworkConnectionResponse, NetworkAddressResponse, NetworkInterfaceStatsResponse, SensorTemperatureResponse, \
    SensorResponseList, SensorFanResponse, SensorBatteryResponse, ConnectedUserResponseList, ConnectUserResponse, \
    ProcessResponseList, ProcessResponse

from platypush.plugins import Plugin, action


class SystemPlugin(Plugin):
    """
    Plugin to get system info.

    Requires:

        - **py-cpuinfo** (``pip install py-cpuinfo``) for CPU model and info.
        - **psutil** (``pip install psutil``) for CPU load and stats.

    """

    @action
    def cpu_info(self) -> CpuInfoResponse:
        """
        Get CPU info.
        :return: :class:`platypush.message.response.system.CpuInfoResponse`
        """
        from cpuinfo import get_cpu_info
        info = get_cpu_info()

        return CpuInfoResponse(
            arch=info.get('raw_arch_string'),
            bits=info.get('bits'),
            count=info.get('count'),
            vendor_id=info.get('vendor_id'),
            brand=info.get('brand'),
            hz_advertised=info.get('hz_advertised_raw')[0],
            hz_actual=info.get('hz_actual_raw')[0],
            stepping=info.get('stepping'),
            model=info.get('model'),
            family=info.get('family'),
            flags=info.get('flags'),
            l1_instruction_cache_size=info.get('l1_instruction_cache_size'),
            l1_data_cache_size=info.get('l1_data_cache_size'),
            l2_cache_size=info.get('l2_cache_size'),
            l3_cache_size=info.get('l3_cache_size'),
        )

    @action
    def cpu_times(self, per_cpu=False, percent=False) -> Union[CpuTimesResponse, CpuResponseList]:
        """
        Get the CPU times stats.

        :param per_cpu: Get per-CPU stats (default: False).
        :param percent: Get the stats in percentage (default: False).
        :return: :class:`platypush.message.response.system.CpuTimesResponse`
        """
        import psutil

        times = psutil.cpu_times_percent(percpu=per_cpu) if percent else \
            psutil.cpu_times(percpu=per_cpu)

        if per_cpu:
            return CpuResponseList([
                CpuTimesResponse(
                    user=t.user,
                    nice=t.nice,
                    system=t.system,
                    idle=t.idle,
                    iowait=t.iowait,
                    irq=t.irq,
                    softirq=t.softirq,
                    steal=t.steal,
                    guest=t.guest,
                    guest_nice=t.guest_nice,
                )
                for t in times
            ])

        return CpuTimesResponse(
            user=times.user,
            nice=times.nice,
            system=times.system,
            idle=times.idle,
            iowait=times.iowait,
            irq=times.irq,
            softirq=times.softirq,
            steal=times.steal,
            guest=times.guest,
            guest_nice=times.guest_nice,
        )

    @action
    def cpu_percent(self, per_cpu: bool = False, interval: Optional[float] = None) -> Union[float, List[float]]:
        """
        Get the CPU load percentage.

        :param per_cpu: Get per-CPU stats (default: False).
        :param interval: When *interval* is 0.0 or None compares system CPU times elapsed since last call or module
            import, returning immediately (non blocking). That means the first time this is called it will
            return a meaningless 0.0 value which you should ignore. In this case is recommended for accuracy that this
            function be called with at least 0.1 seconds between calls.
        :return: float if ``per_cpu=False``, ``list[float]`` otherwise.
        """
        import psutil
        percent = psutil.cpu_percent(percpu=per_cpu, interval=interval)

        if per_cpu:
            return [p for p in percent]
        return percent

    @action
    def cpu_stats(self) -> CpuStatsResponse:
        """
        Get CPU stats.
        :return: :class:`platypush.message.response.system.CpuStatsResponse`
        """
        import psutil
        stats = psutil.cpu_stats()

        return CpuStatsResponse(
            ctx_switches=stats.ctx_switches,
            interrupts=stats.interrupts,
            soft_interrupts=stats.soft_interrupts,
            syscalls=stats.syscalls,
        )

    @action
    def cpu_frequency(self, per_cpu: bool = False) -> Union[CpuFrequencyResponse, CpuResponseList]:
        """
        Get CPU stats.

        :param per_cpu: Get per-CPU stats (default: False).
        :return: :class:`platypush.message.response.system.CpuFrequencyResponse`
        """
        import psutil
        freq = psutil.cpu_freq(percpu=per_cpu)

        if per_cpu:
            return CpuResponseList([
                CpuFrequencyResponse(
                    min=f.min,
                    max=f.max,
                    current=f.current,
                )
                for f in freq
            ])

        return CpuFrequencyResponse(
            min=freq.min,
            max=freq.max,
            current=freq.current,
        )

    @action
    def load_avg(self) -> List[float]:
        """
        Get the average load as a vector that represents the load within the last 1, 5 and 15 minutes.
        """
        import psutil
        return psutil.getloadavg()

    @action
    def mem_virtual(self) -> VirtualMemoryUsageResponse:
        """
        Get the current virtual memory usage stats.
        :return: list of :class:`platypush.message.response.system.VirtualMemoryUsageResponse`
        """
        import psutil
        mem = psutil.virtual_memory()
        return VirtualMemoryUsageResponse(
            total=mem.total,
            available=mem.available,
            percent=mem.percent,
            used=mem.used,
            free=mem.free,
            active=mem.active,
            inactive=mem.inactive,
            buffers=mem.buffers,
            cached=mem.cached,
            shared=mem.shared,
        )

    @action
    def mem_swap(self) -> SwapMemoryUsageResponse:
        """
        Get the current virtual memory usage stats.
        :return: list of :class:`platypush.message.response.system.SwapMemoryUsageResponse`
        """
        import psutil
        mem = psutil.swap_memory()
        return SwapMemoryUsageResponse(
            total=mem.total,
            percent=mem.percent,
            used=mem.used,
            free=mem.free,
            sin=mem.sin,
            sout=mem.sout,
        )

    @action
    def disk_partitions(self) -> DiskResponseList:
        """
        Get the list of partitions mounted on the system.
        :return: list of :class:`platypush.message.response.system.DiskPartitionResponse`
        """
        import psutil
        parts = psutil.disk_partitions()
        return DiskResponseList([
            DiskPartitionResponse(
                device=p.device,
                mount_point=p.mountpoint,
                fstype=p.fstype,
                opts=p.opts,
            ) for p in parts
        ])

    @action
    def disk_usage(self, path: Optional[str] = None) -> Union[DiskUsageResponse, DiskResponseList]:
        """
        Get the usage of a mounted disk.

        :param path: Path where the device is mounted (default: get stats for all mounted devices).
        :return: :class:`platypush.message.response.system.DiskUsageResponse` or list of
            :class:`platypush.message.response.system.DiskUsageResponse`.
        """
        import psutil

        if path:
            usage = psutil.disk_usage(path)
            return DiskUsageResponse(
                 path=path,
                 total=usage.total,
                 used=usage.used,
                 free=usage.free,
                 percent=usage.percent,
            )
        else:
            disks = {p.mountpoint: psutil.disk_usage(p.mountpoint)
                     for p in psutil.disk_partitions()}

            return DiskResponseList([
                DiskUsageResponse(
                    path=path,
                    total=disk.total,
                    used=disk.used,
                    free=disk.free,
                    percent=disk.percent,
                ) for path, disk in disks.items()
            ])

    @action
    def disk_io_counters(self, disk: Optional[str] = None, per_disk: bool = False) -> \
            Union[DiskIoCountersResponse, DiskResponseList]:
        """
        Get the I/O counter stats for the mounted disks.

        :param disk: Select the stats for a specific disk (e.g. 'sda1'). Default: get stats for all mounted disks.
        :param per_disk: Return the stats per disk (default: False).
        :return: :class:`platypush.message.response.system.DiskIoCountersResponse` or list of
            :class:`platypush.message.response.system.DiskIoCountersResponse`.
        """
        import psutil

        def _expand_response(_disk, _stats):
            return DiskIoCountersResponse(
                read_count=_stats.read_count,
                write_count=_stats.write_count,
                read_bytes=_stats.read_bytes,
                write_bytes=_stats.write_bytes,
                read_time=_stats.read_time,
                write_time=_stats.write_time,
                read_merged_count=_stats.read_merged_count,
                write_merged_count=_stats.write_merged_count,
                busy_time=_stats.busy_time,
                disk=_disk,
            )

        if disk:
            per_disk = True

        io = psutil.disk_io_counters(perdisk=per_disk)
        if disk:
            stats = [d for name, d in io.items() if name == disk]
            assert stats, 'No such disk: {}'.format(disk)
            return _expand_response(disk, stats[0])

        if not per_disk:
            return _expand_response(None, io)

        return DiskResponseList([
            _expand_response(disk, stats)
            for disk, stats in io.items()
        ])

    @action
    def net_io_counters(self, nic: Optional[str] = None, per_nic: bool = False) -> \
            Union[NetworkIoCountersResponse, NetworkResponseList]:
        """
        Get the I/O counters stats for the network interfaces.

        :param nic: Select the stats for a specific network device (e.g. 'eth0'). Default: get stats for all NICs.
        :param per_nic: Return the stats broken down per interface (default: False).
        :return: :class:`platypush.message.response.system.NetIoCountersResponse` or list of
            :class:`platypush.message.response.system.NetIoCountersResponse`.
        """
        import psutil

        def _expand_response(_nic, _stats):
            return NetworkIoCountersResponse(
                bytes_sent=_stats.bytes_sent,
                bytes_recv=_stats.bytes_recv,
                packets_sent=_stats.packets_sent,
                packets_recv=_stats.packets_recv,
                errin=_stats.errin,
                errout=_stats.errout,
                dropin=_stats.dropin,
                dropout=_stats.dropout,
                nic=_nic,
            )

        if nic:
            per_nic = True

        io = psutil.net_io_counters(pernic=per_nic)
        if nic:
            stats = [d for name, d in io.items() if name == nic]
            assert stats, 'No such network interface: {}'.format(nic)
            return _expand_response(nic, stats[0])

        if not per_nic:
            return _expand_response(nic, io)

        return NetworkResponseList([
            _expand_response(nic, stats)
            for nic, stats in io.items()
        ])

    # noinspection PyShadowingBuiltins
    @action
    def net_connections(self, type: Optional[str] = None) -> Union[NetworkConnectionResponse, NetworkResponseList]:
        """
        Get the list of active network connections.
        On macOS this function requires root privileges.

        :param type: Connection type to filter. Supported types:

            +------------+----------------------------------------------------+
            | Kind Value | Connections using                                  |
            +------------+----------------------------------------------------+
            | inet       | IPv4 and IPv6                                      |
            | inet4      | IPv4                                               |
            | inet6      | IPv6                                               |
            | tcp        | TCP                                                |
            | tcp4       | TCP over IPv4                                      |
            | tcp6       | TCP over IPv6                                      |
            | udp        | UDP                                                |
            | udp4       | UDP over IPv4                                      |
            | udp6       | UDP over IPv6                                      |
            | unix       | UNIX socket (both UDP and TCP protocols)           |
            | all        | the sum of all the possible families and protocols |
            +------------+----------------------------------------------------+

        :return: List of :class:`platypush.message.response.system.NetworkConnectionResponse`.
        """
        import psutil
        conns = psutil.net_connections(kind=type)

        return NetworkResponseList([
            NetworkConnectionResponse(
                fd=conn.fd,
                family=conn.family.name,
                type=conn.type.name,
                local_address=conn.laddr[0] if conn.laddr else None,
                local_port=conn.laddr[1] if len(conn.laddr) > 1 else None,
                remote_address=conn.raddr[0] if conn.raddr else None,
                remote_port=conn.raddr[1] if len(conn.raddr) > 1 else None,
                status=conn.status,
                pid=conn.pid,
            ) for conn in conns
        ])

    @action
    def net_addresses(self, nic: Optional[str] = None) -> Union[NetworkAddressResponse, NetworkResponseList]:
        """
        Get address info associated to the network interfaces.

        :param nic: Select the stats for a specific network device (e.g. 'eth0'). Default: get stats for all NICs.
        :return: :class:`platypush.message.response.system.NetworkAddressResponse` or list of
            :class:`platypush.message.response.system.NetworkAddressResponse`.
        """
        import psutil
        addrs = psutil.net_if_addrs()

        def _expand_addresses(_nic, _addrs):
            args = {'nic': _nic}

            for addr in _addrs:
                if addr.family == socket.AddressFamily.AF_INET:
                    args.update({
                        'ipv4_address': addr.address,
                        'ipv4_netmask': addr.netmask,
                        'ipv4_broadcast': addr.broadcast,
                    })
                elif addr.family == socket.AddressFamily.AF_INET6:
                    args.update({
                        'ipv6_address': addr.address,
                        'ipv6_netmask': addr.netmask,
                        'ipv6_broadcast': addr.broadcast,
                    })
                elif addr.family == socket.AddressFamily.AF_PACKET:
                    args.update({
                        'mac_address': addr.address,
                        'mac_broadcast': addr.broadcast,
                    })

                if addr.ptp and not args.get('ptp'):
                    args['ptp'] = addr.ptp

            return NetworkAddressResponse(**args)

        if nic:
            addrs = [addr for name, addr in addrs.items() if name == nic]
            assert addrs, 'No such network interface: {}'.format(nic)
            addr = addrs[0]
            return _expand_addresses(nic, addr)

        return NetworkResponseList([
            _expand_addresses(nic, addr)
            for nic, addr in addrs.items()
        ])

    @action
    def net_stats(self, nic: Optional[str] = None) -> Union[NetworkInterfaceStatsResponse, NetworkResponseList]:
        """
        Get stats about the network interfaces.

        :param nic: Select the stats for a specific network device (e.g. 'eth0'). Default: get stats for all NICs.
        :return: :class:`platypush.message.response.system.NetworkInterfaceStatsResponse` or list of
            :class:`platypush.message.response.system.NetworkInterfaceStatsResponse`.
        """
        import psutil
        stats = psutil.net_if_stats()

        def _expand_stats(_nic, _stats):
            return NetworkInterfaceStatsResponse(
                nic=_nic,
                is_up=_stats.isup,
                duplex=_stats.duplex.name,
                speed=_stats.speed,
                mtu=_stats.mtu,
            )

        if nic:
            stats = [addr for name, addr in stats.items() if name == nic]
            assert stats, 'No such network interface: {}'.format(nic)
            return _expand_stats(nic, stats[0])

        return NetworkResponseList([
            _expand_stats(nic, addr)
            for nic, addr in stats.items()
        ])

    # noinspection DuplicatedCode
    @action
    def sensors_temperature(self, sensor: Optional[str] = None, fahrenheit: bool = False) \
            -> Union[SensorTemperatureResponse, List[SensorTemperatureResponse],
                     Dict[str, Union[SensorTemperatureResponse, List[SensorTemperatureResponse]]]]:
        """
        Get stats from the temperature sensors.

        :param sensor: Select the sensor name.
        :param fahrenheit: Return the temperature in Fahrenheit (default: Celsius).
        """
        import psutil
        stats = psutil.sensors_temperatures(fahrenheit=fahrenheit)

        if sensor:
            stats = [addr for name, addr in stats.items() if name == sensor]
            assert stats, 'No such sensor name: {}'.format(sensor)
            if len(stats) == 1:
                return SensorTemperatureResponse(
                    name=sensor,
                    current=stats[0].current,
                    high=stats[0].high,
                    critical=stats[0].critical,
                    label=stats[0].label,
                )

            return [
                SensorTemperatureResponse(
                    name=sensor,
                    current=s.current,
                    high=s.high,
                    critical=s.critical,
                    label=s.label,
                )
                for s in stats
            ]

        ret = {}
        for name, data in stats.items():
            for stat in data:
                resp = SensorTemperatureResponse(
                    name=sensor,
                    current=stat.current,
                    high=stat.high,
                    critical=stat.critical,
                    label=stat.label,
                ).output

                if name not in ret:
                    ret[name] = resp
                else:
                    if isinstance(ret[name], list):
                        ret[name].append(resp)
                    else:
                        ret[name] = [ret[name], resp]

        return ret

    # noinspection DuplicatedCode
    @action
    def sensors_fan(self, sensor: Optional[str] = None) -> SensorResponseList:
        """
        Get stats from the fan sensors.

        :param sensor: Select the sensor name.
        :return: List of :class:`platypush.message.response.system.SensorFanResponse`.
        """
        import psutil
        stats = psutil.sensors_fans()

        def _expand_stats(name, _stats):
            return SensorResponseList([
                SensorFanResponse(
                    name=name,
                    current=s.current,
                    label=s.label,
                )
                for s in _stats
            ])

        if sensor:
            stats = [addr for name, addr in stats.items() if name == sensor]
            assert stats, 'No such sensor name: {}'.format(sensor)
            return _expand_stats(sensor, stats[0])

        return SensorResponseList([
            _expand_stats(name, stat)
            for name, stat in stats.items()
        ])

    @action
    def sensors_battery(self) -> SensorBatteryResponse:
        """
        Get stats from the battery sensor.
        :return: List of :class:`platypush.message.response.system.SensorFanResponse`.
        """
        import psutil
        stats = psutil.sensors_battery()

        return SensorBatteryResponse(
            percent=stats.percent,
            secs_left=stats.secsleft,
            power_plugged=stats.power_plugged,
        )

    @action
    def connected_users(self) -> ConnectedUserResponseList:
        """
        Get the list of connected users.
        :return: List of :class:`platypush.message.response.system.ConnectUserResponse`.
        """
        import psutil
        users = psutil.users()

        return ConnectedUserResponseList([
            ConnectUserResponse(
                name=u.name,
                terminal=u.terminal,
                host=u.host,
                started=datetime.fromtimestamp(u.started),
                pid=u.pid,
            )
            for u in users
        ])

    # noinspection PyShadowingBuiltins
    @action
    def processes(self, filter: Optional[str] = '') -> ProcessResponseList:
        """
        Get the list of running processes.

        :param filter: Filter the list by name.
        :return: List of :class:`platypush.message.response.system.ProcessResponse`.
        """
        import psutil
        processes = [psutil.Process(pid) for pid in psutil.pids()]
        p_list = []

        for p in processes:
            if filter and filter not in p.name():
                continue

            args = {}

            try:
                mem = p.memory_info()
                times = p.cpu_times()
                args.update(
                    pid=p.pid,
                    name=p.name(),
                    started=datetime.fromtimestamp(p.create_time()),
                    ppid=p.ppid(),
                    children=[pp.pid for pp in p.children()],
                    status=p.status(),
                    username=p.username(),
                    terminal=p.terminal(),
                    cpu_user_time=times.user,
                    cpu_system_time=times.system,
                    cpu_children_user_time=times.children_user,
                    cpu_children_system_time=times.children_system,
                    mem_rss=mem.rss,
                    mem_vms=mem.vms,
                    mem_shared=mem.shared,
                    mem_text=mem.text,
                    mem_data=mem.data,
                    mem_lib=mem.lib,
                    mem_dirty=mem.dirty,
                    mem_percent=p.memory_percent(),
                )
            except psutil.Error:
                continue

            try:
                args.update(
                    exe=p.exe(),
                )
            except psutil.Error:
                pass

            p_list.append(ProcessResponse(**args))

        return ProcessResponseList(p_list)

    @staticmethod
    def _get_process(pid: int):
        import psutil
        return psutil.Process(pid)

    @action
    def pid_exists(self, pid: int) -> bool:
        """
        :param pid: Process PID.
        :return: ``True`` if the process exists, ``False`` otherwise.
        """
        import psutil
        return psutil.pid_exists(pid)

    @action
    def suspend(self, pid: int):
        """
        Suspend a process.
        :param pid: Process PID.
        """
        self._get_process(pid).suspend()

    @action
    def resume(self, pid: int):
        """
        Resume a process.
        :param pid: Process PID.
        """
        self._get_process(pid).resume()

    @action
    def terminate(self, pid: int):
        """
        Terminate a process.
        :param pid: Process PID.
        """
        self._get_process(pid).terminate()

    @action
    def kill(self, pid: int):
        """
        Kill a process.
        :param pid: Process PID.
        """
        self._get_process(pid).kill()

    @action
    def wait(self, pid: int, timeout: int = None):
        """
        Wait for a process to terminate.

        :param pid: Process PID.
        :param timeout: Timeout in seconds (default: ``None``).
        """
        self._get_process(pid).wait(timeout)


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend


class SensorBme280Backend(SensorBackend):
    """
    Backend to poll analog sensor values from a `BME280 <https://shop.pimoroni.com/products/bme280-breakout>`_
    environment sensor

    Requires:

        * ``pimoroni-bme280`` (``pip install pimoroni-bme280``)
    """

    def __init__(self, temperature=True, pressure=True, humidity=True, **kwargs):
        """
        :param temperature: Enable temperature sensor polling
        :param pressure: Enable pressure sensor polling
        :param humidity: Enable humidity sensor polling
        """

        enabled_sensors = {
            'temperature': temperature,
            'pressure': pressure,
            'humidity': humidity,
        }

        super().__init__(plugin='gpio.sensor.bme280', enabled_sensors=enabled_sensors, **kwargs)


# vim:sw=4:ts=4:et:



import os

from typing import Optional, Dict, Any, List

from platypush.message.response.printer.cups import PrinterResponse, PrintersResponse, PrinterJobAddedResponse
from platypush.plugins import Plugin, action


class PrinterCupsPlugin(Plugin):
    """
    A plugin to interact with a CUPS printer server.

    Requires:

        - **pycups** (``pip install pycups``)

    """

    def __init__(self, host: str = 'localhost', printer: Optional[str] = None, **kwargs):
        """
        :param host: CUPS host IP/name (default: localhost).
        :param printer: Default printer name that should be used.
        """
        super().__init__(**kwargs)
        self.host = host
        self.printer = printer

    def _get_connection(self, host: Optional[str] = None):
        # noinspection PyPackageRequirements
        import cups
        connection = cups.Connection(host=host or self.host)
        return connection

    def _get_printer(self, printer: Optional[str] = None):
        printer = printer or self.printer
        assert printer, 'No printer specified nor default printer configured'
        return printer

    @action
    def get_printers(self, host: Optional[str] = None) -> PrintersResponse:
        """
        Get the list of printers registered on a CUPS server.
        :param host: CUPS server host IP/name (default: default configured ``host``).
        :return: :class:`platypush.message.response.printer.cups.PrintersResponse`, as a name -> attributes dict.
        """
        conn = self._get_connection(host)
        return PrintersResponse(printers=[
            PrinterResponse(
                name=name,
                printer_type=printer.get('printer-type'),
                info=printer.get('printer-info'),
                uri=printer.get('device-uri'),
                state=printer.get('printer-state'),
                is_shared=printer.get('printer-is-shared'),
                state_message=printer.get('printer-state-message'),
                state_reasons=printer.get('printer-state-reasons', []),
                location=printer.get('printer-location'),
                uri_supported=printer.get('printer-uri-supported'),
                make_and_model=printer.get('printer-make-and-model'),
            )
            for name, printer in conn.getPrinters().items()
        ])

    @action
    def print_test_page(self, printer: Optional[str] = None, host: Optional[str] = None) -> PrinterJobAddedResponse:
        """
        Print the CUPS test page.

        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        job_id = conn.printTestPage(printer)
        return PrinterJobAddedResponse(printer=printer, job_id=job_id)

    @action
    def print_file(self,
                   filename: str,
                   printer: Optional[str] = None,
                   host: Optional[str] = None,
                   title: Optional[str] = None,
                   options: Optional[Dict[str, Any]] = None) -> PrinterJobAddedResponse:
        """
        Print a file.

        :param filename: Path to the file to print.
        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param title: Print title.
        :param options: Extra CUPS name->value options.
        """
        filename = os.path.abspath(os.path.expanduser(filename))
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        job_id = conn.printFile(printer, filename=filename, title=title or '', options=options or {})
        return PrinterJobAddedResponse(printer=printer, job_id=job_id)

    @action
    def print_files(self,
                    filenames: List[str],
                    printer: Optional[str] = None,
                    host: Optional[str] = None,
                    title: Optional[str] = None,
                    options: Optional[Dict[str, Any]] = None) -> PrinterJobAddedResponse:
        """
        Print a list of files.

        :param filenames: Paths to the files to print.
        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param title: Print title.
        :param options: Extra CUPS name->value options.
        """
        filenames = [os.path.abspath(os.path.expanduser(f)) for f in filenames]
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        job_id = conn.printFiles(printer, filenames=filenames, title=title or '', options=options or {})
        return PrinterJobAddedResponse(printer=printer, job_id=job_id)

    @action
    def add_printer(self,
                    name: str,
                    ppd_file: str,
                    info: str,
                    location: Optional[str] = None,
                    host: Optional[str] = None):
        """
        Add a printer.

        :param name: Printer name - alphanumeric + underscore characters only.
        :param ppd_file: Path to the PPD file with the printer information and configuration.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param info: Human-readable information about the printer.
        :param location: Human-readable printer location info.
        """
        conn = self._get_connection(host)
        ppd_file = os.path.abspath(os.path.expanduser(ppd_file))
        # noinspection PyArgumentList
        conn.addPrinter(name=name, filename=ppd_file, info=info, location=location)

    @action
    def delete_printer(self, printer: str, host: Optional[str] = None):
        """
        Delete a printer from a CUPS server.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        conn.deletePrinter(printer)

    @action
    def enable_printer(self, printer: Optional[str], host: Optional[str] = None):
        """
        Enable a printer on a CUPS server.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.enablePrinter(printer)

    @action
    def disable_printer(self, printer: Optional[str] = None, host: Optional[str] = None):
        """
        Disable a printer on a CUPS server.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.disablePrinter(printer)

    @action
    def get_jobs(self, host: Optional[str] = None) -> Dict[int, Dict[str, Any]]:
        """
        Get the list of active jobs.

        :param host: CUPS server IP/name (default: default configured ``host``).
        :return: A job_id -> job_info dict.
        """
        conn = self._get_connection(host)
        return conn.getJobs()

    @action
    def accept_jobs(self, printer: Optional[str] = None, host: Optional[str] = None):
        """
        Start accepting jobs on a printer.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.acceptJobs(printer)

    @action
    def reject_jobs(self, printer: Optional[str] = None, host: Optional[str] = None):
        """
        Start rejecting jobs on a printer.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.rejectJobs(printer)

    @action
    def cancel_job(self, job_id: int, purge_job: bool = False, host: Optional[str] = None):
        """
        Cancel a printer job.

        :param job_id: Job ID to cancel.
        :param purge_job: Also remove the job from the server (default: False).
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        conn.cancelJob(job_id, purge_job=purge_job)

    @action
    def move_job(self,
                 job_id: int,
                 source_printer_uri: str,
                 target_printer_uri: str,
                 host: Optional[str] = None):
        """
        Move a job to another printer/URI.

        :param job_id: Job ID to cancel.
        :param source_printer_uri: Source printer URI.
        :param target_printer_uri: Target printer URI.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        conn.moveJob(printer_uri=source_printer_uri, job_id=job_id, job_printer_uri=target_printer_uri)

    @action
    def finish_document(self, printer: Optional[str] = None, host: Optional[str] = None):
        """
        Finish sending a document to a printer.

        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.finishDocument(printer)

    @action
    def add_printer_to_class(self,
                             printer_class: str,
                             printer: Optional[str] = None,
                             host: Optional[str] = None):
        """
        Add a printer to a class.

        :param printer_class: Class name.
        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.addPrinterToClass(printer, printer_class)

    @action
    def delete_printer_from_class(self,
                                  printer_class: str,
                                  printer: Optional[str] = None,
                                  host: Optional[str] = None):
        """
        Delete a printer from a class.

        :param printer_class: Class name.
        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host)
        printer = self._get_printer(printer)
        conn.deletePrinterFromClass(printer, printer_class)

    @action
    def get_classes(self, host: Optional[str] = None) -> Dict[str, Dict[str, Any]]:
        """
        Get the list of classes on a CUPS server.

        :param host: CUPS server IP/name (default: default configured ``host``).
        :return: dict - class_name -> class_info.
        """
        conn = self._get_connection(host)
        return conn.getClasses()


# vim:sw=4:ts=4:et:



from platypush.plugins import Plugin, action


class MusicPlugin(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @action
    def play(self):
        raise NotImplementedError()

    @action
    def pause(self):
        raise NotImplementedError()

    @action
    def stop(self):
        raise NotImplementedError()

    @action
    def next(self):
        raise NotImplementedError()

    @action
    def previous(self):
        raise NotImplementedError()

    @action
    def setvol(self, vol):
        raise NotImplementedError()

    @action
    def add(self, content):
        raise NotImplementedError()

    @action
    def clear(self):
        raise NotImplementedError()

    @action
    def status(self):
        raise NotImplementedError()


# vim:sw=4:ts=4:et:




from typing import Dict, Any

from platypush.plugins import Plugin, action


class NmapPlugin(Plugin):
    """
    Nmap network scanner/mapper integration.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @action
    def scan(self, hosts: str, ports: str, args: str, sudo: bool = False) -> Dict[str, Any]:
        """
        Perform a port scan towards a certain host or network.

        :param hosts: Host name/IP or IP subnet to scan (e.g. ``192.168.1.0/24``).
        :param ports: Port number, (comma-separated) list or (dash-separated) range to scan (default: all).
        :param args: Additional command line arguments for nmap.
        :param sudo: Execute nmap as root through sudo (default: ``False``).
        :return: Scan results, as an ip -> host map.
        """
        import nmap
        nm = nmap.PortScanner()
        return nm.scan(hosts=hosts, ports=ports, arguments=args, sudo=sudo).get('scan')


# vim:sw=4:ts=4:et:



from typing import List

from platypush.plugins import Plugin, action


class InputsPlugin(Plugin):
    """
    This plugin emulates user input on a keyboard/mouse. It requires the a graphical server (X server or Mac/Win
    interface) to be running - it won't work in console mode.

    Requires:

        * **pyuserinput** (``pip install pyuserinput``)

    """

    @staticmethod
    def _get_keyboard():
        # noinspection PyPackageRequirements
        from pykeyboard import PyKeyboard
        return PyKeyboard()

    @staticmethod
    def _get_mouse():
        # noinspection PyPackageRequirements
        from pymouse import PyMouse
        return PyMouse()

    @classmethod
    def _parse_key(cls, key: str):
        k = cls._get_keyboard()
        keymap = {
            'ctrl': k.control_key,
            'alt': k.alt_key,
            'meta': k.windows_l_key,
            'shift': k.shift_key,
            'enter': k.enter_key,
            'tab': k.tab_key,
            'home': k.home_key,
            'end': k.end_key,
            'capslock': k.caps_lock_key,
            'back': k.backspace_key,
            'del': k.delete_key,
            'up': k.up_key,
            'down': k.down_key,
            'left': k.left_key,
            'right': k.right_key,
            'pageup': k.page_up_key,
            'pagedown': k.page_down_key,
            'esc': k.escape_key,
            'find': k.find_key,
            'f1': k.function_keys[1],
            'f2': k.function_keys[2],
            'f3': k.function_keys[3],
            'f4': k.function_keys[4],
            'f5': k.function_keys[5],
            'f6': k.function_keys[6],
            'f7': k.function_keys[7],
            'f8': k.function_keys[8],
            'f9': k.function_keys[9],
            'f10': k.function_keys[10],
            'f11': k.function_keys[11],
            'f12': k.function_keys[12],
            'help': k.help_key,
            'media_next': k.media_next_track_key,
            'media_prev': k.media_prev_track_key,
            'media_play': k.media_play_pause_key,
            'menu': k.menu_key,
            'numlock': k.num_lock_key,
            'print': k.print_key,
            'print_screen': k.print_screen_key,
            'sleep': k.sleep_key,
            'space': k.space_key,
            'voldown': k.volume_down_key,
            'volup': k.volume_up_key,
            'volmute': k.volume_mute_key,
            'zoom': k.zoom_key,
        }

        lkey = key.lower()
        if lkey in keymap:
            return keymap[lkey]

        return key

    @action
    def press_key(self, key: str):
        """
        Emulate the pressure of a key.
        :param key: Key to be pressed
        """
        kbd = self._get_keyboard()
        key = self._parse_key(key)
        kbd.press_key(key)

    @action
    def release_key(self, key: str):
        """
        Release a pressed key.
        :param key: Key to be released
        """
        kbd = self._get_keyboard()
        key = self._parse_key(key)
        kbd.release_key(key)

    @action
    def press_keys(self, keys: List[str]):
        """
        Emulate the pressure of multiple keys.
        :param keys: List of keys to be pressed.
        """
        kbd = self._get_keyboard()
        keys = [self._parse_key(k) for k in keys]
        kbd.press_keys(keys)

    @action
    def tap_key(self, key: str, repeat: int = 1, interval: float = 0):
        """
        Emulate a key tap.
        :param key: Key to be pressed
        :param repeat: Number of iterations (default: 1)
        :param interval: Repeat interval in seconds (default: 0)
        """
        kbd = self._get_keyboard()
        key = self._parse_key(key)
        kbd.tap_key(key, n=repeat, interval=interval)

    @action
    def type_string(self, string: str, interval: float = 0):
        """
        Type a string.
        :param string: String to be typed
        :param interval: Interval between key strokes in seconds (default: 0)
        """
        kbd = self._get_keyboard()
        kbd.type_string(string, interval=interval)

    @action
    def get_screen_size(self) -> List[int]:
        """
        Get the size of the screen in pixels.
        """
        m = self._get_mouse()
        return [m.screen_size()]

    @action
    def mouse_click(self, x: int, y: int, btn: int, repeat: int = 1):
        """
        Mouse click.
        :param x: x screen position
        :param y: y screen position
        :param btn: Button number (1 for left, 2 for right, 3 for middle)
        :param repeat: Number of clicks (default: 1)
        """
        m = self._get_mouse()
        m.click(x, y, btn, n=repeat)


# vim:sw=4:ts=4:et:



import queue
import threading
from abc import ABC, abstractmethod
from typing import Optional, Union, List

import sounddevice as sd

from platypush.context import get_bus
from platypush.message.event.stt import SpeechDetectionStartedEvent, SpeechDetectionStoppedEvent, SpeechStartedEvent, \
    SpeechDetectedEvent, HotwordDetectedEvent, ConversationDetectedEvent
from platypush.message.response.stt import SpeechDetectedResponse
from platypush.plugins import Plugin, action


class SttPlugin(ABC, Plugin):
    """
    Abstract class for speech-to-text plugins.

    Triggers:

        * :class:`platypush.message.event.stt.SpeechStartedEvent` when speech starts being detected.
        * :class:`platypush.message.event.stt.SpeechDetectedEvent` when speech is detected.
        * :class:`platypush.message.event.stt.SpeechDetectionStartedEvent` when speech detection starts.
        * :class:`platypush.message.event.stt.SpeechDetectionStoppedEvent` when speech detection stops.
        * :class:`platypush.message.event.stt.HotwordDetectedEvent` when a user-defined hotword is detected.
        * :class:`platypush.message.event.stt.ConversationDetectedEvent` when speech is detected after a hotword.

    """

    _thread_stop_timeout = 10.0
    rate = 16000
    channels = 1

    def __init__(self,
                 input_device: Optional[Union[int, str]] = None,
                 hotword: Optional[str] = None,
                 hotwords: Optional[List[str]] = None,
                 conversation_timeout: Optional[float] = 10.0,
                 block_duration: float = 1.0):
        """
        :param input_device: PortAudio device index or name that will be used for recording speech (default: default
            system audio input device).
        :param hotword: When this word is detected, the plugin will trigger a
            :class:`platypush.message.event.stt.HotwordDetectedEvent` instead of a
            :class:`platypush.message.event.stt.SpeechDetectedEvent` event. You can use these events for hooking other
            assistants.
        :param hotwords: Use a list of hotwords instead of a single one.
        :param conversation_timeout: If ``hotword`` or ``hotwords`` are set and ``conversation_timeout`` is set,
            the next speech detected event will trigger a :class:`platypush.message.event.stt.ConversationDetectedEvent`
            instead of a :class:`platypush.message.event.stt.SpeechDetectedEvent` event. You can hook custom hooks
            here to run any logic depending on the detected speech - it can emulate a kind of
            "OK, Google. Turn on the lights" interaction without using an external assistant (default: 10 seconds).
        :param block_duration: Duration of the acquired audio blocks (default: 1 second).
        """

        super().__init__()
        self.input_device = input_device
        self.conversation_timeout = conversation_timeout
        self.block_duration = block_duration

        self.hotwords = set(hotwords or [])
        if hotword:
            self.hotwords = {hotword}

        self._conversation_event = threading.Event()
        self._input_stream: Optional[sd.InputStream] = None
        self._recording_thread: Optional[threading.Thread] = None
        self._detection_thread: Optional[threading.Thread] = None
        self._audio_queue: Optional[queue.Queue] = None
        self._current_text = ''

    def _get_input_device(self, device: Optional[Union[int, str]] = None) -> int:
        """
        Get the index of the input device by index or name.

        :param device: Device index or name. If None is set then the function will return the index of the
            default audio input device.
        :return: Index of the audio input device.
        """
        if not device:
            device = self.input_device
        if not device:
            return sd.query_hostapis()[0].get('default_input_device')

        if isinstance(device, int):
            assert device <= len(sd.query_devices())
            return device

        for i, dev in enumerate(sd.query_devices()):
            if dev['name'] == device:
                return i

        raise AssertionError('Device {} not found'.format(device))

    def on_speech_detected(self, speech: str) -> None:
        """
        Hook called when speech is detected. Triggers the right event depending on the current context.

        :param speech: Detected speech.
        """
        speech = speech.strip()

        if speech in self.hotwords:
            event = HotwordDetectedEvent(hotword=speech)
            if self.conversation_timeout:
                self._conversation_event.set()
                threading.Timer(self.conversation_timeout, lambda: self._conversation_event.clear()).start()
        elif self._conversation_event.is_set():
            event = ConversationDetectedEvent(speech=speech)
        else:
            event = SpeechDetectedEvent(speech=speech)

        get_bus().post(event)

    @staticmethod
    def convert_frames(frames:  bytes) -> bytes:
        """
        Conversion method for raw audio frames. It just returns the input frames as bytes. Override it if required
        by your logic.

        :param frames: Input audio frames, as bytes.
        :return: The audio frames as passed on the input. Override if required.
        """
        return frames

    def on_detection_started(self) -> None:
        """
        Method called when the ``detection_thread`` starts. Initialize your context variables and models here if
        required.
        """
        pass

    def on_detection_ended(self) -> None:
        """
        Method called when the ``detection_thread`` stops. Clean up your context variables and models here.
        """
        pass

    def before_recording(self) -> None:
        """
        Method called when the ``recording_thread`` starts. Put here any logic that you may want to run before the
        recording thread starts.
        """
        pass

    def on_recording_started(self) -> None:
        """
        Method called after the ``recording_thread`` opens the audio device. Put here any logic that you may want to
        run after the recording starts.
        """
        pass

    def on_recording_ended(self) -> None:
        """
        Method called when the ``recording_thread`` stops. Put here any logic that you want to run after the audio
        device is closed.
        """
        pass

    @abstractmethod
    def detect_speech(self, frames) -> str:
        """
        Method called within the ``detection_thread`` when new audio frames have been captured. Must be implemented
        by the derived classes.

        :param frames: Audio frames, as returned by ``convert_frames``.
        :return: Detected text, as a string. Returns an empty string if no text has been detected.
        """
        raise NotImplementedError

    def process_text(self, text: str) -> None:
        if (not text and self._current_text) or (text and text == self._current_text):
            self.on_speech_detected(self._current_text)
            self._current_text = ''
        else:
            if text:
                if not self._current_text:
                    get_bus().post(SpeechStartedEvent())
                self.logger.info('Intermediate speech results: [{}]'.format(text))

            self._current_text = text

    def detection_thread(self) -> None:
        """
        This thread reads frames from ``_audio_queue``, performs the speech-to-text detection and calls
        """
        self._current_text = ''
        self.logger.debug('Detection thread started')
        self.on_detection_started()

        while self._audio_queue:
            try:
                frames = self._audio_queue.get()
                frames = self.convert_frames(frames)
            except Exception as e:
                self.logger.warning('Error while feeding audio to the model: {}'.format(str(e)))
                continue

            text = self.detect_speech(frames).strip()
            self.process_text(text)

        self.on_detection_ended()
        self.logger.debug('Detection thread terminated')

    def recording_thread(self, block_duration: Optional[float] = None, block_size: Optional[int] = None,
                         input_device: Optional[str] = None) -> None:
        """
        Recording thread. It reads raw frames from the audio device and dispatches them to ``detection_thread``.

        :param block_duration: Audio blocks duration. Specify either ``block_duration`` or ``block_size``.
        :param block_size: Size of the audio blocks. Specify either ``block_duration`` or ``block_size``.
        :param input_device: Input device
        """
        assert (block_duration or block_size) and not (block_duration and block_size), \
            'Please specify either block_duration or block_size'

        if not block_size:
            block_size = int(self.rate * self.channels * block_duration)

        self.before_recording()
        self.logger.debug('Recording thread started')
        device = self._get_input_device(input_device)
        self._input_stream = sd.InputStream(samplerate=self.rate, device=device,
                                            channels=self.channels, dtype='int16', latency=0,
                                            blocksize=block_size)
        self._input_stream.start()
        self.on_recording_started()
        get_bus().post(SpeechDetectionStartedEvent())

        while self._input_stream:
            try:
                frames = self._input_stream.read(block_size)[0]
            except Exception as e:
                self.logger.warning('Error while reading from the audio input: {}'.format(str(e)))
                continue

            self._audio_queue.put(frames)

        get_bus().post(SpeechDetectionStoppedEvent())
        self.on_recording_ended()
        self.logger.debug('Recording thread terminated')

    @abstractmethod
    @action
    def detect(self, audio_file: str) -> SpeechDetectedResponse:
        """
        Perform speech-to-text analysis on an audio file. Must be implemented by the derived classes.

        :param audio_file: Path to the audio file.
        """
        raise NotImplementedError

    def __enter__(self):
        """
        Context manager enter. Starts detection and returns self.
        """
        self.start_detection()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager exit. Stops detection.
        """
        self.stop_detection()

    @action
    def start_detection(self, input_device: Optional[str] = None, seconds: Optional[float] = None,
                        block_duration: Optional[float] = None) -> None:
        """
        Start the speech detection engine.

        :param input_device: Audio input device name/index override
        :param seconds: If set, then the detection engine will stop after this many seconds, otherwise it'll
            start running until ``stop_detection`` is called or application stop.
        :param block_duration: ``block_duration`` override.
        """
        assert not self._input_stream and not self._recording_thread, 'Speech detection is already running'
        block_duration = block_duration or self.block_duration
        input_device = input_device if input_device is not None else self.input_device
        self._audio_queue = queue.Queue()
        self._recording_thread = threading.Thread(
            target=lambda: self.recording_thread(block_duration=block_duration, input_device=input_device))

        self._recording_thread.start()
        self._detection_thread = threading.Thread(target=lambda: self.detection_thread())
        self._detection_thread.start()

        if seconds:
            threading.Timer(seconds, lambda: self.stop_detection()).start()

    @action
    def stop_detection(self) -> None:
        """
        Stop the speech detection engine.
        """
        assert self._input_stream, 'Speech detection is not running'
        self._input_stream.stop(ignore_errors=True)
        self._input_stream.close(ignore_errors=True)
        self._input_stream = None

        if self._recording_thread:
            self._recording_thread.join(timeout=self._thread_stop_timeout)
            self._recording_thread = None

        self._audio_queue = None
        if self._detection_thread:
            self._detection_thread.join(timeout=self._thread_stop_timeout)
            self._detection_thread = None


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class SmartCardDetectedEvent(Event):
    """
    Event triggered when a smart card is detected
    """

    def __init__(self, atr, reader=None, *args, **kwargs):
        """
        :param atr: Smart card ATR (Answer To Reset)
        :type atr: str

        :param reader: Name or address of the reader that fired the event
        :type reader: str
        """

        super().__init__(atr=atr, reader=reader, *args, **kwargs)


class SmartCardRemovedEvent(Event):
    """
    Event triggered when a smart card is removed
    """

    def __init__(self, atr=None, reader=None, *args, **kwargs):
        """
        :param atr: Smart card ATR (Answer To Reset)
        :type atr: str

        :param reader: Name or address of the reader that fired the event
        :type reader: str
        """

        super().__init__(atr=atr, reader=reader, *args, **kwargs)


# vim:sw=4:ts=4:et:




import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.weather import NewWeatherConditionEvent


class WeatherDarkskyBackend(Backend):
    """
    Weather forecast backend - listens and propagates new weather events.

    Triggers:

        * :class:`platypush.message.event.weather.NewWeatherConditionEvent` when there is a weather condition update

    Requires:

        * The :class:`platypush.plugins.weather.darksky.WeatherDarkskyPlugin` plugin configured
    """

    def __init__(self, poll_seconds, **kwargs):
        super().__init__(**kwargs)
        self.poll_seconds = poll_seconds
        self.latest_update = {}

    def run(self):
        super().run()
        weather = get_plugin('weather.darksky')
        self.logger.info('Initialized weather forecast backend')

        while not self.should_stop():
            current_weather = weather.get_current_weather().output
            del current_weather['time']

            if current_weather != self.latest_update:
                self.bus.post(NewWeatherConditionEvent(**current_weather))

            self.latest_update = current_weather
            time.sleep(self.poll_seconds)


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.foursquare import FoursquareCheckinEvent


class FoursquareBackend(Backend):
    """
    This backend polls for new check-ins on the user's Foursquare account and triggers an event when a new check-in
    occurs.

    Requires:

        * The :class:`platypush.plugins.foursquare.FoursquarePlugin` plugin configured and enabled.

    Triggers:

        - :class:`platypush.message.event.foursquare.FoursquareCheckinEvent` when a new check-in occurs.

    """

    _last_created_at_varname = '_foursquare_checkin_last_created_at'

    def __init__(self, poll_seconds: Optional[float] = 60.0, *args, **kwargs):
        """
        :param poll_seconds: How often the backend should check for new check-ins (default: one minute).
        """
        super().__init__(*args, poll_seconds=poll_seconds, **kwargs)
        self._last_created_at = None

    def __enter__(self):
        self._last_created_at = int(get_plugin('variable').get(self._last_created_at_varname).
                                    output.get(self._last_created_at_varname) or 0)
        self.logger.info('Started Foursquare backend')

    def loop(self):
        checkins = get_plugin('foursquare').get_checkins().output
        if not checkins:
            return

        last_checkin = checkins[0]
        last_checkin_created_at = last_checkin.get('createdAt', 0)
        if self._last_created_at and last_checkin_created_at <= self._last_created_at:
            return

        self.bus.post(FoursquareCheckinEvent(checkin=last_checkin))
        self._last_created_at = last_checkin_created_at
        get_plugin('variable').set(**{self._last_created_at_varname: self._last_created_at})


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend


class SensorEnvirophatBackend(SensorBackend):
    """
    Backend to poll analog sensor values from an enviroPHAT sensor pHAT
    (https://shop.pimoroni.com/products/enviro-phat)

    Requires:

        * ``envirophat`` (``pip install envirophat``)
    """

    def __init__(self, temperature=True, pressure=True, altitude=True, luminosity=True,
                 analog=True, accelerometer=True, magnetometer=True, qnh=1020, **kwargs):
        """
        :param temperature: Enable temperature sensor polling
        :param pressure: Enable pressure sensor polling
        :param altitude: Enable altitude sensor polling
        :param luminosity: Enable luminosity sensor polling
        :param analog: Enable analog sensors polling
        :param accelerometer: Enable accelerometer polling
        :param magnetometer: Enable magnetometer polling
        :param qnh: Base reference for your sea level pressure (for altitude sensor)
        """

        enabled_sensors = {
            'temperature': temperature,
            'pressure': pressure,
            'altitude': altitude,
            'luminosity': luminosity,
            'analog': analog,
            'accelerometer': accelerometer,
            'magnetometer': magnetometer,
        }

        super().__init__(plugin='gpio.sensor.envirophat',
                         plugin_args={'qnh': qnh},
                         enabled_sensors=enabled_sensors, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Optional, Union

from linode_api4 import LinodeClient, Instance
from platypush.plugins.sensor import SensorPlugin

from platypush.message.response.linode import LinodeInstancesResponse, LinodeInstanceResponse
from platypush.plugins import action


class LinodePlugin(SensorPlugin):
    """
    This plugin can interact with a Linode account and manage node and volumes.

    To get your token:

        - Login to <https://cloud.linode.com/>.
        - Go to My Profile -> API Tokens -> Add a Personal Access Token.
        - Select the scopes that you want to provide to your new token.

    Requires:

        * **linode_api4** (``pip install linode_api4``)

    """

    def __init__(self, token: str, **kwargs):
        """
        :param token: Your Linode token.
        """

        super().__init__(**kwargs)
        self._token = token

    def _get_client(self, token: Optional[str] = None) -> LinodeClient:
        return LinodeClient(token or self._token)

    def _get_instance(self, label: str, token: Optional[str] = None) -> Instance:
        client = self._get_client(token)
        instances = client.linode.instances(Instance.label == label)
        assert instances, 'No such Linode instance: ' + label
        return instances[0]

    @action
    def status(self, token: Optional[str] = None, instance: Optional[str] = None) \
            -> Union[LinodeInstanceResponse, LinodeInstancesResponse]:
        """
        Get the full status and info of the instances associated to a selected account.

        :param token: Override the default access token if you want to query another account.
        :param instance: Select only one node by label.
        :return: :class:`platypush.message.response.linode.LinodeInstanceResponse` if ``label`` is specified,
            :class:`platypush.message.response.linode.LinodeInstancesResponse` otherwise.
        """
        if instance:
            instance = self._get_instance(label=instance)
            return LinodeInstanceResponse(instance=instance)

        client = self._get_client(token)
        return LinodeInstancesResponse(instances=client.linode.instances())

    @action
    def reboot(self, instance: str, token: Optional[str] = None) -> None:
        """
        Reboot an instance.

        :param instance: Label of the instance to be rebooted.
        :param token: Default access token override.
        """
        instance = self._get_instance(label=instance, token=token)
        assert instance.reboot(), 'Reboot failed'

    @action
    def boot(self, instance: str, token: Optional[str] = None) -> None:
        """
        Boot an instance.

        :param instance: Label of the instance to be booted.
        :param token: Default access token override.
        """
        instance = self._get_instance(label=instance, token=token)
        assert instance.boot(), 'Boot failed'

    @action
    def shutdown(self, instance: str, token: Optional[str] = None) -> None:
        """
        Shutdown an instance.

        :param instance: Label of the instance to be shut down.
        :param token: Default access token override.
        """
        instance = self._get_instance(label=instance, token=token)
        assert instance.shutdown(), 'Shutdown failed'

    @action
    def get_measurement(self, *args, **kwargs):
        return self.status(*args, **kwargs)


# vim:sw=4:ts=4:et:



import base64
import json
import serial
import threading
import time

from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


# noinspection PyBroadException
class SerialPlugin(SensorPlugin):
    """
    The serial plugin can read data from a serial device, as long as the serial
    device returns a JSON. You can use this plugin to interact for example with
    some sensors connected through an Arduino. Just make sure that the code on
    your serial device returns JSON values. If you're using an Arduino or any
    ATMega compatible device, take a look at
    https://github.com/bblanchon/ArduinoJson.
    """

    def __init__(self, device=None, baud_rate=9600, **kwargs):
        """
        :param device: Device path (e.g. ``/dev/ttyUSB0`` or ``/dev/ttyACM0``)
        :type device: str

        :param baud_rate: Serial baud rate (default: 9600)
        :type baud_rate: int
        """

        super().__init__(**kwargs)

        self.device = device
        self.baud_rate = baud_rate
        self.serial = None
        self.serial_lock = threading.Lock()
        self.last_measurement = None

    def _read_json(self, serial_port):
        n_brackets = 0
        is_escaped_ch = False
        parse_start = False
        output = bytes()

        while True:
            ch = serial_port.read()
            if not ch:
                break

            try:
                ch.decode()
            except Exception as e:
                self.logger.warning('Could not decode character: {}'.format(str(e)))
                output = bytes()

            if ch.decode() == '{' and not is_escaped_ch:
                parse_start = True
                n_brackets += 1

            if not parse_start:
                continue

            output += ch

            if ch.decode() == '}' and not is_escaped_ch:
                n_brackets -= 1
                if n_brackets == 0:
                    break

            is_escaped_ch = ch.decode() == '\\'

        return output.decode().strip()

    def _get_serial(self, device=None, baud_rate=None, reset=False):
        if not device:
            if not self.device:
                raise RuntimeError('No device specified nor default device configured')
            device = self.device

        if baud_rate is None:
            if self.baud_rate is None:
                raise RuntimeError('No baud_rate specified nor default configured')
            baud_rate = self.baud_rate

        if self.serial:
            if not reset:
                return self.serial

            self._close_serial()

        self.serial = serial.Serial(device, baud_rate)
        return self.serial

    def _close_serial(self):
        if self.serial:
            try:
                self.serial.close()
                self.serial = None
            except Exception as e:
                self.logger.warning('Error while closing serial communication: {}')
                self.logger.exception(e)

    @action
    def get_measurement(self, device=None, baud_rate=None):
        """
        Reads JSON data from the serial device and returns it as a message

        :param device: Device path (default: default configured device)
        :type device: str

        :param baud_rate: Baud rate (default: default configured baud_rate)
        :type baud_rate: int
        """

        if not device:
            if not self.device:
                raise RuntimeError('No device specified nor default device configured')
            device = self.device

        if baud_rate is None:
            if self.baud_rate is None:
                raise RuntimeError('No baud_rate specified nor default configured')
            baud_rate = self.baud_rate

        data = None

        try:
            serial_available = self.serial_lock.acquire(timeout=2)
            if serial_available:
                try:
                    ser = self._get_serial(device=device, baud_rate=baud_rate)
                except:
                    time.sleep(1)
                    ser = self._get_serial(device=device, baud_rate=baud_rate, reset=True)

                data = self._read_json(ser)

                try:
                    data = json.loads(data)
                except:
                    self.logger.warning('Invalid JSON message from {}: {}'.
                                        format(self.device, data))
            else:
                data = self.last_measurement
        finally:
            try:
                self.serial_lock.release()
            except:
                pass

        if data:
            self.last_measurement = data

        return data

    @action
    def read(self, device=None, baud_rate=None, size=None, end=None):
        """
        Reads raw data from the serial device

        :param device: Device to read (default: default configured device)
        :type device: str

        :param baud_rate: Baud rate (default: default configured baud_rate)
        :type baud_rate: int

        :param size: Number of bytes to read
        :type size: int

        :param end: End of message byte or character
        :type end: int, bytes or str
        """

        if not device:
            if not self.device:
                raise RuntimeError('No device specified nor default device configured')
            device = self.device

        if baud_rate is None:
            if self.baud_rate is None:
                raise RuntimeError('No baud_rate specified nor default configured')
            baud_rate = self.baud_rate

        if (size is None and end is None) or (size is not None and end is not None):
            raise RuntimeError('Either size or end must be specified')

        if end and isinstance(end, str) and len(end) > 1:
            raise RuntimeError('The serial end must be a single character, not a string')

        data = bytes()

        try:
            serial_available = self.serial_lock.acquire(timeout=2)
            if serial_available:
                try:
                    ser = self._get_serial(device=device, baud_rate=baud_rate)
                except:
                    time.sleep(1)
                    ser = self._get_serial(device=device, baud_rate=baud_rate, reset=True)

                if size is not None:
                    for _ in range(0, size):
                        data += ser.read()
                elif end is not None:
                    if isinstance(end, str):
                        end = end.encode()

                    ch = None
                    while ch != end:
                        ch = ser.read()

                        if ch != end:
                            data += ch
            else:
                self.logger.warning('Serial read timeout')
        finally:
            try:
                self.serial_lock.release()
            except:
                pass

        try:
            data = data.decode()
        except:
            data = base64.encodebytes(data)

        return data

    @action
    def write(self, data, device=None, baud_rate=None):
        """
        Writes data to the serial device.

        :param device: Device to write (default: default configured device)
        :type device: str

        :param baud_rate: Baud rate (default: default configured baud_rate)
        :type baud_rate: int

        :param data: Data to send to the serial device
        :type data: str, bytes or dict. If dict, it will be serialized as JSON.
        """

        if not device:
            if not self.device:
                raise RuntimeError('No device specified nor default device configured')
            device = self.device

        if baud_rate is None:
            if self.baud_rate is None:
                raise RuntimeError('No baud_rate specified nor default configured')
            baud_rate = self.baud_rate

        if isinstance(data, dict):
            data = json.dumps(data)
        if isinstance(data, str):
            data = data.encode('utf-8')

        try:
            serial_available = self.serial_lock.acquire(timeout=2)
            if serial_available:
                try:
                    ser = self._get_serial(device=device, baud_rate=baud_rate)
                except:
                    time.sleep(1)
                    ser = self._get_serial(device=device, baud_rate=baud_rate, reset=True)

                self.logger.info('Writing {} to {}'.format(data, self.device))
                ser.write(data)
        finally:
            try:
                self.serial_lock.release()
            except:
                pass


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class WidgetUpdateEvent(Event):
    """
    Event delivered to the dashboard when a widget update request is delivered to it
    """

    def __init__(self, widget, *args, **kwargs):
        """
        :param widget: Widget ID
        :type widget: str
        """

        super().__init__(widget=widget, *args, **kwargs)


# vim:sw=4:ts=4:et:




import enum
import logging
import queue
import os
import re
import threading
import websocket
from typing import Optional, Union

from platypush.utils import grouper


class Connection:
    """
    This class models the connection with an ESP8266/ESP32 device over its WebREPL websocket channel.
    """

    _file_transfer_buffer_size = 1024

    class State(enum.IntEnum):
        DISCONNECTED = 1
        CONNECTED = 2
        PASSWORD_REQUIRED = 3
        READY = 4
        SENDING_REQUEST = 5
        WAITING_ECHO = 6
        WAITING_RESPONSE = 7
        SENDING_FILE_TRANSFER_RESPONSE = 8
        WAITING_FILE_TRANSFER_RESPONSE = 9
        UPLOADING_FILE = 10
        DOWNLOADING_FILE = 11

    class FileRequestType(enum.IntEnum):
        UPLOAD = 1
        DOWNLOAD = 2

    def __init__(self, host: str, port: int, connect_timeout: Optional[float] = None,
                 password: Optional[str] = None, ws: Optional[websocket.WebSocketApp] = None):
        self.host = host
        self.port = port
        self.connect_timeout = connect_timeout
        self.password = password
        self.state = self.State.DISCONNECTED
        self.ws = ws
        self._connected = threading.Event()
        self._logged_in = threading.Event()
        self._echo_received = threading.Event()
        self._response_received = threading.Event()
        self._download_chunk_ready = threading.Event()
        self._file_transfer_request_ack_received = threading.Event()
        self._file_transfer_ack_received = threading.Event()
        self._file_transfer_request_successful = True
        self._file_transfer_successful = True
        self._downloaded_chunks = queue.Queue()
        self._password_requested = False
        self._running_cmd = None
        self._received_echo = None
        self._received_response = None
        self._paste_header_received = False
        self.logger = logging.getLogger(__name__)

    def send(self, msg: Union[str, bytes], wait_response: bool = True, timeout: Optional[float] = None):
        bufsize = 255

        msg = (msg
               .replace("\n", "\r\n")  # end of command in normal mode
               .replace("\\x01", "\x01")  # switch to raw mode
               .replace("\\x02", "\x02")  # switch to normal mode
               .replace("\\x03", "\x03")  # interrupt
               .replace("\\x04", "\x04")  # end of command in raw mode
               .encode())

        if not msg.endswith(b'\r\n'):
            msg += b'\r\n'
        if wait_response:
            # Enter PASTE mode and exit on end-of-message
            msg = b'\x05' + msg + b'\x04'

        if wait_response:
            self.state = self.State.SENDING_REQUEST
            self._running_cmd = msg.decode().strip()
            self._received_echo = ''
            self._response_received.clear()

        self._echo_received.clear()
        for chunk in grouper(bufsize, msg):
            self.ws.send(bytes(chunk))

        if not wait_response:
            return

        self.state = self.State.WAITING_ECHO
        echo_received = self._echo_received.wait(timeout=timeout)
        if not echo_received:
            self.on_timeout('No response echo received')

        self._paste_header_received = False
        response_received = self._response_received.wait(timeout=timeout)
        if not response_received:
            self.on_timeout('No response received')

        response = self._received_response
        self._received_response = None
        return response

    def on_connect(self):
        self.state = Connection.State.CONNECTED
        self._connected.set()

    def on_password_requested(self):
        self._password_requested = True
        self.state = Connection.State.PASSWORD_REQUIRED
        assert self.password, 'This device is protected by password and no password was provided'
        self.send(self.password, wait_response=False)

    def on_ready(self):
        self.state = Connection.State.READY
        self._logged_in.set()

    def on_close(self):
        self.state = self.State.DISCONNECTED
        self._connected.clear()
        self._logged_in.clear()
        self._password_requested = False
        self.ws = None

    def on_recv_echo(self, echo):
        def str_transform(s: str):
            s = s.replace('\x05', '').replace('\x04', '').replace('\r', '')
            s = re.sub('^[\s\r\n]+', '', s)
            s = re.sub('[\s\r\n]+$', '', s)
            return s

        if echo.endswith('\r\n=== ') and not self._paste_header_received:
            self._paste_header_received = True
            return

        if re.match('\s+>>>\s+', echo) \
                or re.match('\s+\.\.\.\s+', echo) \
                or re.match('\s+===\s+', echo):
            return

        self._received_echo += echo
        running_cmd = str_transform(self._running_cmd)
        received_echo = str_transform(self._received_echo)

        if running_cmd == received_echo:
            self._received_echo = None
            self.state = self.State.WAITING_RESPONSE
            self._echo_received.set()

    def close(self):
        # noinspection PyBroadException
        try:
            self.ws.close()
        except:
            pass

        self.on_close()

    def on_timeout(self, msg: str = ''):
        self.close()
        raise TimeoutError(msg)

    def append_response(self, response):
        if isinstance(response, bytes):
            response = response.decode()

        if not self._received_response:
            self._received_response = ''

        self._received_response += response

    def on_recv_response(self, response):
        self.append_response(response)
        self.state = self.State.READY
        self._received_response = self._received_response.strip()

        if self._received_response.startswith('=== '):
            # Strip PASTE mode output residual
            self._received_response = self._received_response[4:]
            self._received_response = self._received_response.strip()

        # Replace \r\n serial end-of-line with \n
        self._received_response = self._received_response.replace('\r\n', '\n')

        # Notify the listeners
        self._response_received.set()

    def on_recv_file_transfer_response(self, response):
        self._file_transfer_request_successful = self._parse_file_transfer_response(response)
        self._file_transfer_request_ack_received.set()

    def on_file_transfer_completed(self, response):
        self._file_transfer_successful = self._parse_file_transfer_response(response)
        self._file_transfer_ack_received.set()

    def on_file_upload_start(self):
        self.logger.info('Starting file upload')
        self._file_transfer_successful = False
        self._file_transfer_ack_received.clear()
        self.state = self.State.UPLOADING_FILE

    def on_file_download_start(self):
        self.logger.info('Starting file download')
        self._file_transfer_successful = False
        self._downloaded_chunks = queue.Queue()
        self.state = self.State.DOWNLOADING_FILE
        self._file_transfer_ack_received.clear()
        self.ws.send(b'\x00', opcode=websocket.ABNF.OPCODE_BINARY)

    def on_chunk_received(self, data):
        size = data[0] | (data[1] << 8)
        data = data[2:]
        if len(data) != size:
            return

        self.logger.info('Received chunk of size {} (total size={})'.format(len(data), size))
        if size == 0:
            # End of file
            self.logger.info('File download completed')
            self._downloaded_chunks.put(None)
            self.on_file_download_completed()
        else:
            self._downloaded_chunks.put(data)
            self.ws.send(b'\x00', opcode=websocket.ABNF.OPCODE_BINARY)

        self._download_chunk_ready.set()
        self._download_chunk_ready.clear()

    def on_file_download_completed(self):
        self.state = self.State.READY

    def on_file_transfer_request(self):
        self.logger.info('Sending file transfer request')
        self._file_transfer_request_successful = False
        self._file_transfer_request_ack_received.clear()
        self.state = self.State.SENDING_FILE_TRANSFER_RESPONSE

    def get_downloaded_chunks(self, timeout: Optional[float] = None):
        while True:
            try:
                chunk = self._downloaded_chunks.get(timeout=timeout)
            except queue.Empty:
                self.on_timeout('File download timed out')

            if chunk is None:
                break

            yield chunk

    def wait_ready(self):
        connected = self._connected.wait(timeout=self.connect_timeout)
        if not connected:
            self.on_timeout('Connection timed out')

        logged_in = self._logged_in.wait(timeout=self.connect_timeout)
        if not logged_in:
            self.on_timeout('Log in timed out')

    def wait_file_request_ack_received(self, timeout):
        self.state = self.State.WAITING_FILE_TRANSFER_RESPONSE
        self._file_transfer_request_ack_received.wait(timeout=timeout)
        assert self._file_transfer_request_successful, 'File transfer request failed'
        self.logger.info('File transfer request acknowledged')

    def wait_file_transfer_completed(self, timeout):
        self._file_transfer_ack_received.wait(timeout)
        assert self._file_transfer_successful, 'File transfer failed'
        self.logger.info('File transfer completed')

    @staticmethod
    def _parse_file_transfer_response(response: bytes) -> bool:
        if not response or len(response) < 4:
            return False

        if response[0] == ord('W') and response[1] == ord('B'):
            return response[2] | response[3] << 8 == 0
        return False

    def _send_file_request(self, filename: str, request_type: FileRequestType, file_size: int = 0,
                           timeout: Optional[float] = None):
        self.on_file_transfer_request()

        # 2 + 1 + 1 + 8 + 4 + 2 + 64
        request = bytearray(82)

        # Protocol mode (file transfer) and request type (1=PUT, 2=GET)
        request[0] = ord('W')
        request[1] = ord('A')
        request[2] = request_type.value

        # File size encoding
        request[12] = file_size & 0xff
        request[13] = (file_size >> 8) & 0xff
        request[14] = (file_size >> 16) & 0xff
        request[15] = (file_size >> 24) & 0xff

        # File name length encoding
        request[16] = len(filename) & 0xff
        request[17] = (len(filename) >> 8) & 0xff

        # File name encoding
        for i in range(0, min(64, len(filename))):
            request[i+18] = ord(filename[i])

        self.ws.send(request, opcode=websocket.ABNF.OPCODE_BINARY)
        self.wait_file_request_ack_received(timeout=timeout)

    def _upload_file(self, f, timeout):
        self.on_file_upload_start()
        content = f.read(self._file_transfer_buffer_size)

        while content:
            self.ws.send(content, opcode=websocket.ABNF.OPCODE_BINARY)
            content = f.read(self._file_transfer_buffer_size)

        self.wait_file_transfer_completed(timeout=timeout)
        self.state = self.State.READY

    def _download_file(self, f, timeout):
        self.on_file_download_start()
        for chunk in self.get_downloaded_chunks(timeout=timeout):
            f.write(chunk)

        self.on_file_download_completed()

    def file_upload(self, source, destination, timeout):
        source = os.path.abspath(os.path.expanduser(source))
        destination = os.path.join(destination, os.path.basename(source)) if destination else os.path.basename(source)
        size = os.path.getsize(source)

        with open(source, 'rb') as f:
            self._send_file_request(destination, self.FileRequestType.UPLOAD, file_size=size, timeout=timeout)
            self._upload_file(f, timeout=timeout)

    def file_download(self, file, fd, timeout=None):
        self._send_file_request(file, request_type=self.FileRequestType.DOWNLOAD, timeout=timeout)
        self._download_file(fd, timeout=timeout)


# vim:sw=4:ts=4:et:



import inspect
import os
import subprocess
import sys

from platypush.backend import Backend


class NoderedBackend(Backend):
    """
    This backend publishes platypush actions to a Node-RED instance.
    If you enable this backend on a host that runs Node-RED then a new block (platypush -> run) can be
    used in your flows. This block will accept JSON requests as input in the format
    ``{"type":"request", "action":"plugin.name.action_name", "args": {...}}`` and return the output
    of the action as block output, or raise an exception if the action failed.

    Requires:

        * **pynodered** (``pip install pynodered``)
    """

    def __init__(self, port: int = 5051, *args, **kwargs):
        """
        :param port: Listening port for the local publishing web server (default: 5051)
        """
        super().__init__(*args, **kwargs)
        self.port = port
        self._runner_path = os.path.join(
            os.path.dirname(inspect.getfile(self.__class__)), 'runner.py')
        self._server = None

    def on_stop(self):
        if self._server:
            self._server.terminate()
            self._server = None

    def run(self):
        super().run()

        self._server = subprocess.Popen([sys.executable, '-m', 'pynodered.server',
                                         '--port', str(self.port), self._runner_path])

        self.logger.info('Started Node-RED backend on port {}'.format(self.port))
        self._server.wait()


# vim:sw=4:ts=4:et:



import base64
from typing import Optional, List

from pyzbar.pyzbar import Decoded
from pyzbar.locations import Rect

from platypush.message import Mapping
from platypush.message.response import Response


class QrcodeResponse(Response):
    pass


class QrcodeGeneratedResponse(QrcodeResponse):
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 content: str,
                 format: str,
                 data: Optional[str] = None,
                 image_file: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(*args, output={
            'text': content,
            'data': data,
            'format': format,
            'image_file': image_file,
        }, **kwargs)


class RectModel(Mapping):
    def __init__(self, rect: Rect):
        super().__init__()
        self.left = rect.left
        self.top = rect.top
        self.width = rect.width
        self.height = rect.height


class ResultModel(Mapping):
    def __init__(self, result: Decoded, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # noinspection PyBroadException
        try:
            data = result.data.decode()
        except:
            data = base64.encodebytes(result.data).decode()

        self.data = data
        self.type = result.type
        self.rect = dict(RectModel(result.rect)) if result.rect else {}


class QrcodeDecodedResponse(QrcodeResponse):
    def __init__(self, results: List[Decoded], image_file: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, output={
            'image_file': image_file,
            'results': [dict(ResultModel(result)) for result in results],
        }, **kwargs)


# vim:sw=4:ts=4:et:



import gzip
import os
import requests
import tempfile
import threading

from platypush.plugins import Plugin, action
from platypush.utils import find_files_by_ext


class MediaSubtitlesPlugin(Plugin):
    """
    Plugin to get video subtitles from OpenSubtitles

    Requires:

        * **python-opensubtitles** (``pip install -e 'git+https://github.com/agonzalezro/python-opensubtitles#egg=python-opensubtitles'``)
        * **webvtt** (``pip install webvtt-py``), optional, to convert srt subtitles into vtt format ready for web streaming
        * **requests** (``pip install requests``)
    """

    def __init__(self, username, password, language=None, **kwargs):
        """
        :param username: Your OpenSubtitles username
        :type username: str

        :param password: Your OpenSubtitles password
        :type password: str

        :param language: Preferred language name, ISO639 code or OpenSubtitles
            language ID to be used for the subtitles. Also supports an (ordered)
            list of preferred languages
        :type language: str or list[str]
        """

        from pythonopensubtitles.opensubtitles import OpenSubtitles

        super().__init__(**kwargs)

        self._ost = OpenSubtitles()
        self._token = self._ost.login(username, password)
        self.languages = []
        self._file_lock = threading.RLock()

        if language:
            if isinstance(language, str):
                self.languages.append(language.lower())
            elif isinstance(language, list):
                self.languages.extend([l.lower() for l in language])
            else:
                raise AttributeError('{} is neither a string nor a list'.format(
                    language))

    @action
    def get_subtitles(self, resource, language=None):
        """
        Get the subtitles data for a video resource

        :param resource: Media file, torrent or URL to the media resource
        :type resource: str

        :param language: Language name or code (default: configured preferred language).
            Choose 'all' for all the languages
        :type language: str
        """

        from pythonopensubtitles.utils import File

        if resource.startswith('file://'):
            resource = resource[len('file://'):]

        resource = os.path.abspath(os.path.expanduser(resource))
        if not os.path.isfile(resource):
            return None, '{} is not a valid file'.format(resource)

        file = resource
        cwd = os.getcwd()
        media_dir = os.path.dirname(resource)
        os.chdir(media_dir)
        file = file.split(os.sep)[-1]

        local_subs = [{
            'IsLocal': True,
            'MovieName': '[Local subtitle]',
            'SubFileName': sub.split(os.sep)[-1],
            'SubDownloadLink': 'file://' + os.path.join(media_dir, sub),
        } for sub in find_files_by_ext(media_dir, '.srt', '.vtt')]

        self.logger.info('Found {} local subtitles for {}'.format(
            len(local_subs), file))

        languages = [language.lower()] if language else self.languages

        try:
            file_hash = File(file).get_hash()
            subs = self._ost.search_subtitles([{
                'sublanguageid': 'all',
                'moviehash': file_hash,
            }])

            subs = [
                sub for sub in subs if not languages or languages[0] == 'all' or
                sub.get('LanguageName', '').lower() in languages or
                sub.get('SubLanguageID', '').lower() in languages or
                sub.get('ISO639', '').lower() in languages
            ]

            for sub in subs:
                sub['IsLocal'] = False

            self.logger.info('Found {} OpenSubtitles items for {}'.format(
                len(subs), file))

            return local_subs + subs
        finally:
            os.chdir(cwd)

    @action
    def download(self, link, media_resource=None, path=None, convert_to_vtt=False):
        """
        Downloads a subtitle link (.srt/.vtt file or gzip/zip OpenSubtitles archive link) to the specified directory

        :param link: Local subtitles file or OpenSubtitles gzip download link
        :type link: str

        :param path: Path where the subtitle file will be downloaded (default: temporary file under /tmp)
        :type path: str

        :param media_resource: Name of the media resource. If set and if it's a
            media local file then the subtitles will be saved in the same folder
        :type media_resource: str

        :param convert_to_vtt: If set to True, then the downloaded subtitles
            will be converted to VTT format (default: no conversion)
        :type convert_to_vtt: bool

        :returns: dict.

        Format::

            {
                "filename": "/path/to/subtitle/file.srt"
            }

        """

        if link.startswith('file://'):
            link = link[len('file://'):]
        if os.path.isfile(link):
            if convert_to_vtt:
                link = self.to_vtt(link).output
            return {'filename': link}

        gzip_content = requests.get(link).content

        if not path and media_resource:
            if media_resource.startswith('file://'):
                media_resource = media_resource[len('file://'):]
            if os.path.isfile(media_resource):
                media_resource = os.path.abspath(media_resource)
                path = os.path.join(
                    os.path.dirname(media_resource),
                    '.'.join(os.path.basename(media_resource).split('.')[:-1])) + '.srt'

        if path:
            f = open(path, 'wb')
        else:
            f = tempfile.NamedTemporaryFile(prefix='media_subs_',
                                            suffix='.srt', delete=False)
            path = f.name

        try:
            with f:
                f.write(gzip.decompress(gzip_content))
            if convert_to_vtt:
                path = self.to_vtt(path).output
        except Exception as e:
            os.unlink(path)
            raise e

        return {'filename': path}

    @action
    def to_vtt(self, filename):
        """
        Get the VTT content given an SRT file. Will return the original content if
        the file is already in VTT format.
        """

        if filename.lower().endswith('.vtt'):
            return filename

        import webvtt

        with self._file_lock:
            try:
                webvtt.read(filename)
                return filename
            except Exception:
                webvtt.from_srt(filename).save()
                return '.'.join(filename.split('.')[:-1]) + '.vtt'


# vim:sw=4:ts=4:et:
